/*
	Combined Multiple Recursive Random Number Generator.

	This implementation is borrowed entirely from the C code in

	L'Ecuyer, Pierre., "Good parameters and implementations for combined
	multiple recursive random number generators,"

	(http://www.iro.umontreal.ca/~lecuyer/myftp/papers/combmrg2.ps)

*/


#include "Python.h"
#ifdef NUMERIC
#include "arrayobject.h"
#endif
#include <math.h>

#define NORM 2.328306549295728e-10
#define M1   4294967087.0
#define M2   4294944443.0
#define A12     1403580.0
#define A13N     810728.0
#define A21      527612.0
#define A23N    1370589.0

typedef struct {
	PyObject_HEAD
	double series[6];
} cmrg32Object;

staticforward PyTypeObject Cmrg32Type;
	
static PyObject *ErrorObject;

static void ErrorSet(char *msg)
{
	if (!ErrorObject)
		ErrorObject = PyString_FromString("cmrg32Error");
	PyErr_SetString(ErrorObject, msg);
}

static cmrg32Object * new_cmrg32Object(double s1, double s2, double s3,
                                       double s4, double s5, double s6)
{
	cmrg32Object * self;

	/* check that not all seeds of each set are zero */

	if ( ( (s1==0.0e0) && (s2==0.0e0) && (s3==0.0e0) ) ||
             ( (s4==0.0e0) && (s5==0.0e0) && (s6==0.0e0) ) ) {
		ErrorSet("All seeds are zero\n");
		return NULL;
	}

	/* check that the first set of seeds are in the range 0 to M1 */
	if ((s1<0.0e0 || s1>=M1) || (s2<0.0e0 || s2>=M1) || (s3<0.0e0 || s3>=M1)) {
		ErrorSet("First three seeds must be in range 1 to 4294967087");
		return NULL;
	}

	/* check that the second set of seeds are in the range 0 to M1 */
	if ((s4<0.0e0 || s4>=M2) || (s5<0.0e0 || s5>=M2) || (s6<0.0e0 || s6>=M2)) {
		ErrorSet("First three seeds must be in range 1 to 4294944443");
		return NULL;
	}

	self = PyObject_NEW(cmrg32Object, &Cmrg32Type);
	if (self == NULL)
		return NULL;

	/* we need to ensure that the seeds are exactly integral */

	self->series[0] = floor(s1);
	self->series[1] = floor(s2);
	self->series[2] = floor(s3);
	self->series[3] = floor(s4);
	self->series[4] = floor(s5);
	self->series[5] = floor(s6);
	return self;
}
static double cmrg_random(cmrg32Object *self)
{
	long   k;
	double p1, p2;
	double *s;

	s = self->series;

	p1 = A12 * s[1] - A13N * s[0];
	k = p1 / M1;
	p1 -= k * M1;
	if (p1 < 0.0)
		p1 += M1;
	s[0] = s[1];
	s[1] = s[2];
	s[2] = p1;

	p2 = A21 * s[5] - A23N * s[3];
	k  = p2 / M2;
	p2 -= k * M2;
	if (p2 < 0.0)
		p2 += M2;
	s[3] = s[4];
	s[4] = s[5];
	s[5] = p2;

	if (p1 <= p2)
		return ((p1 - p2 + M1) * NORM);
	else
		return ((p1 - p2) * NORM);
}

static PyObject * cmrg32_cmrg32(PyObject *self, PyObject *args)
{
	double s1, s2, s3, s4, s5, s6;
	PyObject *result;

	if (!PyArg_ParseTuple(args, "dddddd", &s1, &s2, &s3, &s4, &s5, &s6))
		return NULL;

	result = (PyObject *)new_cmrg32Object(s1, s2, s3, s4, s5, s6);
	return result;
}

static struct PyMethodDef cmrg32_module_methods[] = {
	{"generator",(PyCFunction)cmrg32_cmrg32, METH_VARARGS, NULL},
	{NULL, (PyCFunction)NULL, 0, NULL}
};

static int cmrg_savestate(cmrg32Object *self, char *filename)
{
	FILE *ofilep;
	char buf[255];
	int i;

	ofilep = fopen(filename, "w");
	if (ofilep == NULL) {
		sprintf(buf, "%s: Error Opening File for writing", filename);
		ErrorSet(buf);
		return 0;
	}

	for (i=0; i<6; i++) {
		if (fprintf(ofilep, "%f\n", self->series[i]) < 0) {
			ErrorSet("Error encountered during save");
			goto fail;
		}
	}

	if (fclose(ofilep) != 0) {
		ErrorSet("Error encountered while closing file");
		return 0;
	}
	return 1;
 fail:
	if (fclose(ofilep) != 0) {
		ErrorSet("Error encountered while closing file");
		return 0;
	}
	return 0;
}

static int cmrg_getstate(cmrg32Object *self, char *filename)
{
	FILE *ofilep;
	char buf[255];
	double s[6], x;
	int i;

	ofilep = fopen(filename, "r");
	if (ofilep == NULL) {
		sprintf(buf, "%s: Error Opening File for reading", filename);
		ErrorSet(buf);
		return 0;
	}

	for (i=0; i<6; i++) {
		if (fscanf(ofilep, "%lf", &x) < 0) {
			ErrorSet("Error encountered during read");
			goto fail;
		}
		s[i] = x;
	}

	if (fclose(ofilep) != 0) {
		ErrorSet("Error encountered while closing file");
		return 0;
	}

	for (i=0; i<6; i++)
		self->series[i] = s[i];

	return 1;
 fail:
	if (fclose(ofilep) != 0) {
		ErrorSet("Error encountered while closing file");
		return 0;
	}
	return 0;
}

static PyObject * cmrg32_get_seed(cmrg32Object *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
	return Py_BuildValue("(dddddd)", self->series[0], \
                                         self->series[1], \
                                         self->series[2], \
                                         self->series[3], \
                                         self->series[4], \
                                         self->series[5]);
}
static PyObject * cmrg32_random(cmrg32Object *self, PyObject *args)
{

	if (!PyArg_ParseTuple(args, ""))
		return NULL;

	return PyFloat_FromDouble(cmrg_random(self));
}

static PyObject * cmrg32_sample(cmrg32Object *self, PyObject *args)
{
	PyObject *result, *tmp;
	double ranval;
	int n, i;

	if (!PyArg_ParseTuple(args, "i", &n))
		return NULL;

	result = PyList_New(n);
	if (result == NULL)
		return NULL;

	for (i=0; i<n; i++) {
		ranval = cmrg_random(self);
		tmp = PyFloat_FromDouble(ranval);
		PyList_SetItem(result, i, tmp);
	}
	return result;
}

#ifdef NUMERIC
static PyObject * cmrg32_asample(cmrg32Object *self, PyObject *args)
{
	PyObject *result, *tmp;
	double ranval, *array;
	int n, i;

	if (!PyArg_ParseTuple(args, "i", &n))
		return NULL;

	result = PyArray_FromDims(1, &n, PyArray_DOUBLE);
	if (result == NULL)
		return NULL;

	array = (double *)(((PyArrayObject *)result)->data);

	for (i=0; i<n; i++) {
		array[i] = cmrg_random(self);
	}
	return result;
}
#endif /*NUMERIC */

static PyObject *cmrg32_choice(cmrg32Object *self, PyObject *args)
{
	PyObject *seq, *result;
	double ranval;
	int    size, index;

	if (!PyArg_ParseTuple(args, "O", &seq))
		return NULL;
	if (!PySequence_Check(seq)) {
		ErrorSet("argument to choice must be sequence type");
		return NULL;
	}

	size = PySequence_Length(seq);
	if (size <=0 ) {
		Py_INCREF(Py_None);
		return Py_None;
	}

	ranval = cmrg_random(self);
	index = (int )(size * ranval);
	result = PySequence_GetItem(seq, index);
	return result;
}

static PyObject * cmrg32_uniform(cmrg32Object *self, PyObject *args)
{
	PyObject *ostart;
	PyObject *oend;
	double start, end, diff;
/* 	PyObject *tmp1=NULL; */
/* 	PyObject *tmp2=NULL; */
/* 	PyObject *tmp3=NULL; */
/* 	PyObject *result=NULL; */
	double ranval;

	if (!(PyTuple_GET_SIZE(args) == 2)) {
		ErrorSet("uniform requires two floats as arguments");
		return NULL;
	}
	ostart = PyTuple_GET_ITEM(args, 0);
	oend = PyTuple_GET_ITEM(args, 1);
	if (PyFloat_Check(ostart)) {
		start = PyFloat_AS_DOUBLE(ostart);
	}
	else {
		ErrorSet("uniform requires two floats as arguments");
		return NULL;
	}

	if (PyFloat_Check(oend)){
		end = PyFloat_AS_DOUBLE(oend);
	}
	else {
		ErrorSet("uniform requires two floats as arguments");
		return NULL;
	}
	diff = end - start;
	ranval = cmrg_random(self);
	return PyFloat_FromDouble(start + ranval*diff);

/* 	if (!PyArg_ParseTuple(args, "OO", &start, &end)) */
/* 		return NULL; */
/* 	if (!(PyNumber_Check(start) && PyNumber_Check(end))) { */
/* 		ErrorSet("Arguments to uniform must be numbers"); */
/* 		return NULL; */
/* 	} */

/* 	ranval = cmrg_random(self); */

/* 	if ((tmp1 = PyNumber_Subtract(end, start)) == NULL) */
/* 		goto fail; */
/* 	if ((tmp2 = PyFloat_FromDouble(ranval)) == NULL) */
/* 		goto fail; */
/* 	if((tmp3 = PyNumber_Multiply(tmp1, tmp2)) == NULL) */
/* 		goto fail; */
/* 	if ((result = PyNumber_Add(start, tmp3)) == NULL) */
/* 		goto fail; */
/* 	Py_DECREF(tmp1); */
/* 	Py_DECREF(tmp2); */
/* 	Py_DECREF(tmp3); */
/* 	return result; */
/*   fail: */
/* 	Py_XDECREF(tmp1); */
/* 	Py_XDECREF(tmp2); */
/* 	Py_XDECREF(tmp3); */
/* 	return NULL; */
}

static PyObject * cmrg32_randint(cmrg32Object *self, PyObject *args)
{
	int start, end;
	double ranval, res;

	if (!PyArg_ParseTuple(args, "ii", &start, &end))
		return NULL;

	ranval = cmrg_random(self);
	res = ranval * ((double )end - (double )start + 1.0e0);
	res += (double )start;
	return PyInt_FromLong((long )res);
}

static PyObject * cmrg32_permutation(cmrg32Object *self, PyObject *args)
{
	PyObject *seq;
	double ranval;
	PyObject *tmp1, *tmp2;
	PyObject *result;
	int n, i, pos;

	if (!PyArg_ParseTuple(args, "O", &seq))
		return NULL;

	if (!PySequence_Check(seq)) {
		ErrorSet("Argument to permutation must be sequence type");
		return NULL;
	}

	result = PySequence_List(seq);
	if (result == NULL) {
		ErrorSet("Out of memory in permutation");
		return NULL;
	}
	n = PySequence_Length(seq);
	for (i=0; i<n; i++) {
		ranval = cmrg_random(self);
		pos = (int ) (ranval * (i+1));
		tmp1 = PyList_GetItem(result, i);
		Py_INCREF(tmp1);
		tmp2 = PyList_GetItem(result, pos);
		Py_INCREF(tmp2);
		PyList_SetItem(result, i, tmp2);
		PyList_SetItem(result, pos, tmp1);
	}
	return result;
}

static PyObject * cmrg32_savestate(cmrg32Object *self, PyObject *args)
{
	char *fname;

	if (!PyArg_ParseTuple(args, "s", &fname))
		return NULL;

	if (!cmrg_savestate(self, fname))
		return NULL;
	else
		return PyInt_FromLong(1L);
}

static PyObject * cmrg32_getstate(cmrg32Object *self, PyObject *args)
{
	char *fname;

	if (!PyArg_ParseTuple(args, "s", &fname))
		return NULL;

	if (!cmrg_getstate(self, fname))
		return NULL;
	else
		return PyInt_FromLong(1L);
}

static struct PyMethodDef cmrg32_object_methods[] = {
	{"get_seed",   (PyCFunction)cmrg32_get_seed,    METH_VARARGS, NULL},
	{"random",     (PyCFunction)cmrg32_random,      METH_VARARGS, NULL},
	{"sample",     (PyCFunction)cmrg32_sample,      METH_VARARGS, NULL},
#ifdef NUMERIC
	{"asample",    (PyCFunction)cmrg32_asample,     METH_VARARGS, NULL},
#endif
	{"choice",     (PyCFunction)cmrg32_choice,      METH_VARARGS, NULL},
	{"uniform",    (PyCFunction)cmrg32_uniform,     METH_VARARGS, NULL},
	{"randint",    (PyCFunction)cmrg32_randint,     METH_VARARGS, NULL},
	{"permutation",(PyCFunction)cmrg32_permutation, METH_VARARGS, NULL},
	{"savestate",  (PyCFunction)cmrg32_savestate,   METH_VARARGS, NULL},
	{"getstate",   (PyCFunction)cmrg32_getstate,    METH_VARARGS, NULL},
	{NULL, (PyCFunction)NULL, 0, NULL}
};

static void
cmrg32_dealloc(cmrg32Object *self)
{
	PyMem_DEL(self);
}

static PyObject * cmrg32_getattr(cmrg32Object *self, char *name)
{
	return Py_FindMethod(cmrg32_object_methods, (PyObject *)self, name);
}

static PyObject * cmrg32_repr(cmrg32Object *self)
{
	char buf[] = "cmrg32Object";
	return PyString_FromString(buf);
}

static int * cmrg32_print(cmrg32Object *self, FILE *fp, int flags)
{
	fprintf(fp, "%s", "cmrg32Object");
	return 0;
}

static char Cmrg32Type__doc__[] =
""
;
statichere PyTypeObject Cmrg32Type = {
	PyObject_HEAD_INIT(NULL)
	0,
	"cmrg32Object",
	sizeof(cmrg32Object),
	0,
	(destructor)cmrg32_dealloc, /*tp_dealloc*/
	(printfunc)cmrg32_print,     /*tp_print*/
	(getattrfunc)cmrg32_getattr,	      /*tp_getattr*/
	0,		              /*tp_setattr*/
	0,                            /*tp_compare*/
	(reprfunc)cmrg32_repr,      /*tp_repr*/
	0,	                      /*tp_as_number*/
	0,		              /*tp_as_sequence*/
	0,		              /*tp_as_mapping*/
	0,                            /*tp_hash*/
	0,
	(reprfunc)cmrg32_repr,
	0,
	0,
	0,
	0,
	Cmrg32Type__doc__
};

void initcmrg32()
{
	PyObject *m;
#ifdef NUMERIC
#ifndef _ARRAY_MODULE
import_array();
#endif
#endif
	Cmrg32Type.ob_type = &PyType_Type;
	m = Py_InitModule4("cmrg32", cmrg32_module_methods,
			   NULL, NULL, PYTHON_API_VERSION);
	if (PyErr_Occurred())
		Py_FatalError("can't initialize module cmrg32");
}
