"""This module defines a class Atom3d for representing an atom in a molecule
in LINUS and the following functions

    o *angle* - angle, in degrees, between 3 atoms

    o *torsion* - torsion angle, in degreees, between 4 atoms

    o *ztox* - compute cartesian coordinates from internal coordinates
    for a collection of atoms

    o *commit_internal_coords* - save internal coordinates follwing a
    conformation change

    o *restore_internal_coords* - restore the conformation of a collection
    of atoms following rejection of a conformation change

    o *radius_of_gyration* - radius of gyration of a collection of atoms

    o *geometric_center* - centroid of a collection of atoms, assuming
    unit mass for all atoms

    o *bond_rotate* - rotate a collection of atoms by a specified amount
    about a specified bond
"""    
import math

RADIANS_TO_DEGREES = 180.0/math.pi
DEGREES_TO_RADIANS = math.pi/180.0

class Atom3d:
    """Class to describe an atom in a molecule

    Attributes

        o *name* - string - name of an atom in PDB format

        o *radius* - float - hard sphere radius of an atom

        o *x*, *y*, *z* - float - cartesian coordinates of atom

        o *first_parent*, *second_parent*, *third_parent* - instances of
        Atom3d - Z-matrix description of an atom

        o *fp_distance*, *sp_angle*, *tp_torsion* - float - distance,
        angle in degrees and torsion angle in degrees made by atom with its
        first, second and third parents

        o *resnum* - integer - index of residue that atom is a part of

    """
    def __init__(self):
        self.name = ''
        self.radius = 0.0
        self.x = self.y = self.z = 0.0
        self.first_parent = self.second_parent = self.third_parent = None
        self.fp_distance = self.sp_angle = self.tp_torsion = 0.0
        self.fp_distance_tmp = self.sp_angle_tmp = self.tp_torsion_tmp = 0.0
        self.resnum = 0

    def __hash__(self):
        return hash(repr(self.resnum) + '_' + self.name)

    def coords(self):
        """Cartesian coordinates of an atom as a list"""
        return [self.x, self.y, self.z]

    def distance(self, other, sqrt=math.sqrt):
        """Euclidean distance between two atoms"""
        return sqrt((self.x - other.x)**2.0 + (self.y - other.y)**2.0 +
                    (self.z - other.z)**2.0)

    def sqdistance(self, other):
        """square of the Euclidean distance between two atoms"""

        return (self.x - other.x)**2.0 + (self.y - other.y)**2.0 + \
               (self.z - other.z)**2.0

    def close(self, other, dist=None):
        """check if two atoms are within a specified distance"""

        if dist is None:
            r1 = self.radius
            r2 = other.radius
            if r1 < 1.0e-6 or r2 < 1.0e-6: return 0
            dist = r1 + r2
        d2 = dist * dist
        dx = (self.x - other.x)**2.0
        if dx >= d2: return 0
        dy = (self.y - other.y)**2.0
        if dy >= d2: return 0
        dz = (self.z - other.z)**2.0
        if dz >= d2: return 0
        return dx + dy + dz < d2

    def int_to_cart(self, sin=math.sin, cos=math.cos, sqrt=math.sqrt,
                        abs=math.fabs):
        """generate cartesian coordinates for an atom from it's
        internal coordinates"""

        
        a2 = self.first_parent
        if a2 is None:
            self.x = self.y = self.z = 0.0
            return

        a3 = self.second_parent
        if a3 is None:
            self.x = self.fp_distance
            self.y = self.z = 0.0
            return


        ang = self.sp_angle * DEGREES_TO_RADIANS
        sina = sin(ang)
        cosa = -cos(ang)

        a4 = self.third_parent
        if a4 is None:
            self.z = 0.0
            self.x = a2.x + self.fp_distance*cosa
            self.y = self.fp_distance * sina
            return

        tor = self.tp_torsion * DEGREES_TO_RADIANS
        sint = sina * sin(tor)
        cost = sina * cos(tor)

        u1x = a3.x - a4.x
        u1y = a3.y - a4.y
        u1z = a3.z - a4.z
        d = 1.0 / sqrt(u1x**2.0 + u1y**2.0 + u1z**2.0)
        u1x = u1x * d
        u1y = u1y * d
        u1z = u1z * d

        u2x = a2.x - a3.x
        u2y = a2.y - a3.y
        u2z = a2.z - a3.z
        d = 1.0 / sqrt(u2x**2.0 + u2y**2.0 + u2z**2.0)
        u2x = u2x * d
        u2y = u2y * d
        u2z = u2z * d

        cosine = u1x*u2x + u1y*u2y + u1z*u2z

        if abs(cosine) < 1.0:
            sine = 1.0/sqrt(1.0 - cosine*cosine)
        else:
            sine = 1.0/sqrt(cosine*cosine - 1.0)
            
        u3x = sine * (u1y*u2z - u1z*u2y)
        u3y = sine * (u1z*u2x - u1x*u2z)
        u3z = sine * (u1x*u2y - u1y*u2x)

        u4x = cost * (u3y*u2z - u3z*u2y)
        u4y = cost * (u3z*u2x - u3x*u2z)
        u4z = cost * (u3x*u2y - u3y*u2x)

        dist = self.fp_distance
        self.x = a2.x + dist*(u2x*cosa + u4x + u3x*sint)
        self.y = a2.y + dist*(u2y*cosa + u4y + u3y*sint)
        self.z = a2.z + dist*(u2z*cosa + u4z + u3z*sint)

    def commit_internal_coords(self):
        """Save internal coordinates following acceptance of a conformation
        change"""
        self.fp_distance_tmp = self.fp_distance
        self.sp_angle_tmp = self.sp_angle
        self.tp_torsion_tmp = self.tp_torsion

    def restore_internal_coords(self):
        """Restore internal coordinates following rejection of a conformation
        change"""
        self.fp_distance = self.fp_distance_tmp
        self.sp_angle = self.sp_angle_tmp
        self.tp_torsion = self.tp_torsion_tmp


def angle(a1, a2, a3, acos=math.acos, sqrt=math.sqrt):
    """Angle in degrees between 3 atoms"""
    x1 = a1.x - a2.x
    y1 = a1.y - a2.y
    z1 = a1.z - a2.z

    x2 = a3.x - a2.x
    y2 = a3.y - a2.y
    z2 = a3.z - a2.z

    try:
        ang = (x1*x2 + y1*y2 + z1*z2)/sqrt((x1**2.0 + y1**2.0 + z1**2.0) *
                                           (x2**2.0 + y2**2.0 + z2**2.0))
    except ZeroDivisionError:
        return -999.999
    
    try:
        return acos(ang)*RADIANS_TO_DEGREES
    except ValueError:
        if ang < -1.0:
            return 180.0
        else:
            return 0.0

def torsion(a1, a2, a3, a4, sqrt=math.sqrt, acos=math.acos):
    """Torsion angle in degrees between 4 atoms"""
    a = a1.x - a2.x; b = a1.y - a2.y; c = a1.z - a2.z;
    d = a3.x - a2.x; e = a3.y - a2.y; f = a3.z - a2.z;
    g = a4.x - a3.x; h = a4.y - a3.y; i = a4.z - a3.z;

    ax = b*f - e*c; ay = c*d - f*a; az = a*e - b*d;
    bx = h*f - e*i; by = i*d - f*g; bz = g*e - h*d;

    try:
        ang = (ax*bx+ay*by+az*bz)/sqrt((ax**2.0 + ay**2.0 + az**2.0) *
                                       (bx**2.0 + by**2.0 + bz**2.0))
    except ZeroDivisionError:
        return -999.999
    else:
        try:
            ang = acos(ang) * RADIANS_TO_DEGREES
        except ValueError:
            if ang < -1.0:
                return 180.0
            else:
                return 0.0
        else:
            if ax*(e*bz-f*by) + ay*(f*bx-d*bz) + az*(d*by-e*bx) >= 0.0:
                return -ang
            else:
                return ang

def ztox(atoms, start=None, end=None):
    if start is None: start = 0
    if end is None: end = len(atoms)
    for i in range(start, end):
        atoms[i].int_to_cart()
    

def commit_internal_coords(atoms, start=None, end=None):
    if start is None: start = 0
    if end is None: end = len(atoms)
    for i in range(start, end):
        atoms[i].commit_internal_coords()

def restore_internal_coords(atoms, start=None, end=None):
    if start is None: start = 0
    if end is None: end = len(atoms)
    for i in range(start, end):
        atoms[i].restore_internal_coords()

def radius_of_gyration(atoms):
    rg = x = y = z = 0.0
    for atom in atoms:
        x = x + atom.x
        y = y + atom.y
        z = z + atom.z
    n = 1.0/len(atoms)
    x = x*n
    y = y*n
    z = z*n
    for atom in atoms:
        rg = rg + (atom.x - x)**2.0 + (atom.y - y)**2.0 + (atom.z - z)**2.0
    return math.sqrt(rg*n)

def geometric_center(atoms):
    n = 1.0/len(atoms)
    x = y = z = 0.0
    for atom in atoms:
        x = x + atom.x
        y = y + atom.y
        z = z + atom.z
    return x*n, y*n, z*n


def bond_rotate(b1, b2, atoms, theta):
    t = theta * DEGREES_TO_RADIANS
    x0 = b1.x
    y0 = b1.y
    z0 = b1.z
    d = b1.distance(b2)
    Lambda = (b2.x - x0)/dist
    mu = (b2.y - y0)/dist
    nu = (b2.z - z0)/dist
    sinth = sin(t)
    costh = cos(t)
    zip = 1.0 - costh
    rot11=costh+Lambda*Lambda*zip
    rot12=Lambda*mu*zip-nu*sinth
    rot13=Lambda*nu*zip+mu*sinth
    rot21=Lambda*mu*zip+nu*sinth
    rot22=costh+mu*mu*zip
    rot23=mu*nu*zip-Lambda*sinth
    rot31=Lambda*nu*zip-mu*sinth
    rot32=mu*nu*zip+Lambda*sinth
    rot33=costh+nu*nu*zip
    for atom in atoms:
        dx = atom.x - x0
        dy = atom.y - y0
        dz = atom.z - z0
        atom.x = x0 + dx*rot11 + dy*rot12 + dz*rot13
        atom.y = y0 + dx*rot21 + dy*rot22 + dz*rot23
        atom.z = z0 + dx*rot31 + dy*rot32 + dz*rot33

    
           
try:
    from Atom3dc import *
except ImportError:
    print 'Using slower Python version of Atom3d'
    pass
