"""This module contains procedures for modifying the conformation of a
protein chain.

Data Exported

   o *GLY_COIL* - the possible coil values for glycine residues.
     This is a tuple of tuples, where each tuple lists the
     range of phi, psi and omega respectively

   o *PRO_COIL* - possible coil values for proline

   o *NPG_COIL* - possible coil values for all residues that are not
     proline or glycine.

   o *PRO_PRO_TURN* - possible turn values for two successive proline
     residues.  This is a tuple of tuples. Each individual tuple is of
     length 2, with the first item being a tuple of length 3 listing
     the range of the phi, psi and omega values for the residue at
     position i+1 of a beta-turn and the second item listing the
     values for the i+2 residue.

   o *PRO_X_TURN* - possible turn values for  two successive residues
     in which the first residue is a proline

   o *X_X_TURN* - possible turn values for two successive residues
     where neither is a proline

Functions exported

   o *set_helix* - set the conformation of a specified segement to
     helical conformation

   o *set_strand* -  set the conformation of a specified segement to
     strand conformation
 
   o *set_PII* -  set the conformation of a specified segement to
     PII conformation
 
   o *set_coil* - set the conformation of a specified segement to
     coil conformation

   o *set_turn* - set the conformation of a specified segement to
     turn conformation

   o *set_mesostate_conformation* - set the conformation of a residue
     by specifying a mesostate

   o *choose_conformation* - choose a possible move for a residue
     based on it's weights and specified mesostates

"""     

     
GLY_COIL = (
            ((  70.0,  90.0), (-180.0,-160.0), (175.0, 185.0)),
            (( -70.0, -90.0), ( 125.0, 145.0), (175.0, 185.0)),
            ((  90.0, 120.0), ( 165.0, 195.0), (175.0, 185.0)),
            (( 165.0, 195.0), ( -15.0,  15.0), (175.0, 185.0)),
            (( 105.0, 135.0), ( -15.0,  15.0), (175.0, 185.0)),
            (( -85.0, -55.0), ( 150.0, 180.0), (175.0, 185.0)),
            ((-135.0,-105.0), ( -20.0,  10.0), (175.0, 185.0)),
            ((  50.0,  80.0), (-150.0,-120.0), (175.0, 185.0)),
            (( -90.0, -60.0), ( -25.0,   5.0), (175.0, 185.0))
           )

PRO_COIL = (
            (( -70.0, -50.0), ( -60.0, -30.0), (175.0, 185.0)),
            (( -80.0, -60.0), ( -35.0,  -5.0), (175.0, 185.0)),
            (( -75.0, -55.0), ( 150.0, 180.0), (175.0, 185.0)),
            (( -70.0, -50.0), ( 125.0, 155.0), (175.0, 185.0)),
            (( -70.0, -50.0), ( -60.0, -30.0), ( -5.0,   5.0)),
            (( -80.0, -60.0), ( -35.0,  -5.0), ( -5.0,   5.0)),
            (( -75.0, -55.0), ( 150.0, 180.0), ( -5.0,   5.0)),
            (( -70.0, -50.0), ( 125.0, 155.0), ( -5.0,   5.0))
           )

NPG_COIL = (
            ((-180.0,-150.0), ( -30.0, +30.0), (175.0, 185.0)), # A
            ((-150.0, -90.0), ( -30.0, +30.0), (175.0, 185.0)), # G
            (( -90.0, -60.0), (   0.0, +30.0), (175.0, 185.0)), # lowG+
            ((-180.0,-150.0), ( -90.0, -30.0), (175.0, 185.0)), # F
            ((-150.0, -90.0), ( -90.0, -30.0), (175.0, 185.0)), # L
            (( -90.0, -20.0), ( -90.0,  -0.0), (175.0, 185.0)), # R+
            ((-180.0,-150.0), (-150.0, -90.0), (175.0, 185.0)), # E
            ((-150.0, -90.0), (-150.0, -90.0), (175.0, 185.0)), # K
            (( -90.0, -60.0), (-150.0, -90.0), (175.0, 185.0)), # halfQ
            ((-150.0, -90.0), ( 150.0,-150.0), (175.0, 185.0)), # J
            (( -90.0, -60.0), ( 150.0,-150.0), (175.0, 185.0)), # P
            ((-150.0, -90.0), (  90.0, 150.0), (175.0, 185.0)), # I
            (( -90.0, -60.0), (  90.0, 150.0), (175.0, 185.0)), # O
            ((  30.0,  90.0), (-150.0, -90.0), (175.0, 185.0)), # o
            ((  30.0,  90.0), ( 180.0,-150.0), (175.0, 185.0)), # halfp
            ((  60.0,  90.0), ( 100.0, 180.0), (175.0, 185.0)), # halfpq
            ((  40.0,  80.0), (  20.0,  80.0), (175.0, 185.0)), # partr
            ((-150.0, -90.0), (  30.0,  90.0), (175.0, 185.0))  # H
           )

PRO_PRO_TURN = (
                 (( -75.0, -55.0), ( 138.0, 168.0), ( 175.0, 185.0),
                  ( -99.0, -69.0), ( 155.0, 185.0), ( 175.0, 185.0)),
                 (( -75.0, -55.0), ( -60.0, -30.0), (  -5.0,   5.0),
                  ( -75.0, -55.0), (-155.0,-125.0), ( 175.0, 185.0))
               )
PRO_X_TURN = (
                 (( -75.0, -55.0), ( 138.0, 168.0), ( 175.0, 185.0),
                  ( -99.0, -69.0), ( 155.0, 185.0), ( 175.0, 185.0)),
                 (( -70.0, -50.0), ( -60.0, -30.0), ( 175.0, 185.0),
                  (  65.0,  95.0), ( 170.0, 210.0), ( 175.0, 185.0))
             )
X_PRO_TURN = (
                 (( -75.0, -55.0), ( 138.0, 168.0), ( 175.0, 185.0),
                  ( -99.0, -69.0), ( 155.0, 185.0), ( 175.0, 185.0)),
                 (( -75.0, -55.0), ( -60.0, -30.0), (  -5.0,   5.0),
                  ( -75.0, -55.0), (-155.0,-125.0), ( 175.0, 185.0)),
                 ((-170.0,-130.0), ( -60.0, -30.0), (  -5.0,   5.0),
                  ( -75.0, -55.0), ( -60.0, -30.0), ( 175.0, 185.0))
             )
X_X_TURN = (
                 (( -75.0, -55.0), ( 138.0, 168.0), ( 175.0, 185.0),
                  ( -99.0, -69.0), ( 155.0, 185.0), ( 175.0, 185.0)),
                 (( -70.0, -50.0), ( -60.0, -30.0), ( 175.0, 185.0),
                  (  65.0,  95.0), ( 170.0, 210.0), ( 175.0, 185.0)),
                 ((  55.0,  75.0), (-168.0,-138.0), ( 175.0, 185.0),
                  (  69.0,  99.0), ( 175.0, 205.0), ( 175.0, 185.0)),
                 ((  50.0,  70.0), (  30.0,  60.0), ( 175.0, 185.0),
                  ( -95.0, -65.0), ( 165.0, 195.0), ( 175.0, 185.0))
           )

X_X_TURN_MS = ('OP', 'PP',
               'Ro', 'Rp',
               'oo', 'op', 'pp',
               'rP', 'rJ')

STRAND_MS = 'LG'
PRO_STRAND_MS = 'MR'
GLY_STRAND_MS = 'LGAlg'

from chidict import CHI_VALUES
from mesodict import MESO_VALUES

def set_helix(p, start, end, generator):
    """Set a segment to helical conformation

    Arguments

        o *p* - instance of linusMol - the protein whose conformation is
          to be changed
          
        o *start* - integer - index of the first residue of the segment
          whose conformation is to be changed

        o *end* - integer - index of the last residue of the segment whose
          conformation is to be changed. **Note** only the conformation
          of residues *start* to *end-1* will be changed

        o *generator* - object - a random number generator
        

    Description

        For each residue in the specified segment the conformation is
        set to a helical conformation, i.e. phi in the range -78 to -48,
        psi in the range -57 to -27 and omega in the range 180 +/- 5.
        The conformation of a residue is changed if and only if it
        has a non-zero helical weight.  Chi-values for the residues
        are chosen at random.  This function modifies the Z-matrix of
        the molecule.  Cartesian coordinates are *not* calculated.
    """    

    uniform = generator.uniform
    phi_atoms = p.phi_atoms
    psi_atoms = p.psi_atoms
    ome_atoms = p.omega_atoms
    chi_atoms = p.chi_atoms
    hw = p.residue_helix_weights
    rn = p.residue_names

    for i in range(start, end):
        if hw[i] < 1.0e-9: continue
        phi = uniform( -78.0, -48.0)
        psi = uniform( 123.0, 153.0)
        ome = uniform( 175.0, 185.0)
        if ome > 180.0: ome = ome - 360.0

        try:
            phi_atoms[i].tp_torsion = phi
        except TypeError:
            pass
        else:
            phi_atoms[i].sp_angle = uniform(111.0, 113.0)

        try:
            psi_atoms[i].tp_torsion = psi
        except TypeError:
            pass

        try:
            ome_atoms[i].tp_torsion = ome
        except TypeError:
            pass

        chi_atom = chi_atoms[i]
        if chi_atom:
            values = generator.choice(CHI_VALUES[rn[i]])
            pos = 0
            for atom in chi_atom:
                chi = apply(uniform, values[pos])
                if chi > 180.0: chi = chi - 360.0
                elif chi < -180.0: chi = chi + 360.0
                atom.tp_torsion = chi
                pos = pos + 1

def set_strand(p, start, end, generator):
    """Set a segment to strand conformation

    Arguments

        o *p* - instance of linusMol - the protein whose conformation is
          to be changed
          
        o *start* - integer - index of the first residue of the segment
          whose conformation is to be changed

        o *end* - integer - index of the last residue of the segment whose
          conformation is to be changed. **Note** only the conformation
          of residues *start* to *end-1* will be changed

        o *generator* - object - a random number generator
        

    Description

        For each residue in the specified segment the conformation is
        set to a strand conformation, i.e. phi in the range -150 to -120,
        psi in the range 130 to 160 and omega in the range 180 +/- 5.
        The conformation of a residue is changed if and only if it
        has a non-zero strand weight.  Chi-values for the residues
        are chosen at random.  This function modifies the Z-matrix of
        the molecule.  Cartesian coordinates are *not* calculated.
    """    

    uniform = generator.uniform
    randint = generator.randint
    phi_atoms = p.phi_atoms
    psi_atoms = p.psi_atoms
    ome_atoms = p.omega_atoms
    chi_atoms = p.chi_atoms
    hw = p.residue_strand_weights
    rn = p.residue_names

    for i in range(start, end):
        if hw[i] < 1.0e-9: continue
        if rn[i] <> 'PRO':
            phi = uniform(-150.0,-120.0)
        else:
            phi = -70.0
        psi = uniform( -50.0, -20.0)
        ome = uniform( 175.0, 185.0)
        if ome > 180.0: ome = ome - 360.0

        try:
            phi_atoms[i].tp_torsion = phi
        except TypeError:
            pass
        else:
            phi_atoms[i].sp_angle = uniform(108.0, 111.0)

        try:
            psi_atoms[i].tp_torsion = psi
        except TypeError:
            pass

        try:
             ome_atoms[i].tp_torsion = ome
        except TypeError:
            pass

        chi_atom = chi_atoms[i]
        if chi_atom:
            values = generator.choice(CHI_VALUES[rn[i]])
            pos = 0
            for atom in chi_atom:
                chi = apply(uniform, values[pos])
                if chi > 180.0: chi = chi - 360.0
                elif chi < -180.0: chi = chi + 360.0
                atom.tp_torsion = chi
                pos = pos + 1

def set_PII(p, start, end, generator):
    """Set a segment to PII conformation

    Arguments

        o *p* - instance of linusMol - the protein whose conformation is
          to be changed
          
        o *start* - integer - index of the first residue of the segment
          whose conformation is to be changed

        o *end* - integer - index of the last residue of the segment whose
          conformation is to be changed. **Note** only the conformation
          of residues *start* to *end-1* will be changed

        o *generator* - object - a random number generator
        

    Description

        For each residue in the specified segment the conformation is
        set to a PII conformation, i.e. phi in the range -100 to -20,
        psi in the range 120 to 170 and omega in the range 180 +/- 5.
        The conformation of a residue is changed if and only if it
        has a non-zero PII weight.  Chi-values for the residues
        are chosen at random.  This function modifies the Z-matrix of
        the molecule.  Cartesian coordinates are *not* calculated.
    """    

    uniform = generator.uniform
    randint = generator.randint
    phi_atoms = p.phi_atoms
    psi_atoms = p.psi_atoms
    ome_atoms = p.omega_atoms
    chi_atoms = p.chi_atoms
    hw = p.residue_PII_weights
    rn = p.residue_names

    for i in range(start, end):
        if hw[i] < 1.0e-9: continue
        phi = uniform(-100.0,-20.0)
        psi = uniform( -60.0, -10.0)
        ome = uniform( 175.0, 185.0)
        if ome > 180.0: ome = ome - 360.0

        try:
            phi_atoms[i].tp_torsion = phi
        except TypeError:
            pass
        else:
            phi_atoms[i].sp_angle = uniform(108.0, 111.0)

        try:
            psi_atoms[i].tp_torsion = psi
        except TypeError:
            pass

        try:
             ome_atoms[i].tp_torsion = ome
        except TypeError:
            pass

        chi_atom = chi_atoms[i]
        if chi_atom:
            values = generator.choice(CHI_VALUES[rn[i]])
            pos = 0
            for atom in chi_atom:
                chi = apply(uniform, values[pos])
                if chi > 180.0: chi = chi - 360.0
                elif chi < -180.0: chi = chi + 360.0
                atom.tp_torsion = chi
                pos = pos + 1

def set_super_helix(p, start, end, generator):

    uniform = generator.uniform
    phi_atoms = p.phi_atoms
    psi_atoms = p.psi_atoms
    ome_atoms = p.omega_atoms
    chi_atoms = p.chi_atoms
    hw = p.residue_helix_weights

    for i in range(start, end):
        if hw[i] < 1.0e-9: continue
        phi = uniform( -69.0, -59.0)
        psi = uniform( 132.0, 142.0)
        ome = uniform( 179.0, 181.0)
        if ome > 180.0: ome = ome - 360.0

        try:
            phi_atoms[i].tp_torsion = phi
        except TypeError:
            pass
        else:
            phi_atoms[i].sp_angle = uniform(111.0, 113.0)

        try:
            psi_atoms[i].tp_torsion = psi
        except TypeError:
            pass

        try:
            ome_atoms[i].tp_torsion = ome
        except TypeError:
            pass

        chi_atom = chi_atoms[i]
        if chi_atom:
            chi = generator.choice((-60.0, 180.0))
            chi_atom.tp_torsion = chi

def set_coil(p, start, end, generator):
    """Set a segment to coil conformation

    Arguments

        o *p* - instance of linusMol - the protein whose conformation is
          to be changed
          
        o *start* - integer - index of the first residue of the segment
          whose conformation is to be changed

        o *end* - integer - index of the last residue of the segment whose
          conformation is to be changed. **Note** only the conformation
          of residues *start* to *end-1* will be changed

        o *generator* - object - a random number generator
        

    Description

        For each residue in the specified segment the conformation is
        set to a coil conformation.  The choice of phi, psi, omega
        values depends on the type of residue (GLY, PRO or other).
        Chi-values for the residues are chosen at random.
        This function modifies the Z-matrix of
        the molecule.  Cartesian coordinates are *not* calculated.
    """    

    uniform = generator.uniform
    choice  = generator.choice
    phi_atoms = p.phi_atoms
    psi_atoms = p.psi_atoms
    ome_atoms = p.omega_atoms
    chi_atoms = p.chi_atoms
    rn = p.residue_names
    cw = p.residue_coil_weights

    for i in range(start, end):
        if cw[i] <= 1.0e-9: continue
        if rn[i] == 'PRO':
            f, y, w = choice(PRO_COIL)
        elif rn[i] == 'GLY':
            f, y, w = choice(GLY_COIL)
        else:
            f, y, w = choice(NPG_COIL)

        phi = apply(uniform, f)
        if phi > 180.0:
            phi = 360.0 - phi
        elif phi < -180.0:
            phi = phi + 360.0

        psi = apply(uniform, y)
        if psi > 180.0:
            psi = 360.0 - psi
        elif psi < -180.0:
            psi = psi + 360.0
            
        ome = apply(uniform, w)
        if ome > 180.0:
            ome = 360.0 - ome
        elif ome < -180.0:
            ome = ome + 360.0

        try:
            phi_atoms[i].tp_torsion = phi
        except TypeError:
            pass
        else:
            phi_atoms[i].sp_angle = uniform(108.0, 113.0)

        try:
            psi_atoms[i].tp_torsion = psi
        except TypeError:
            pass

        try:
             ome_atoms[i].tp_torsion = ome
        except TypeError:
            pass

        chi_atom = chi_atoms[i]
        if chi_atom:
            values = choice(CHI_VALUES[rn[i]])
            pos = 0
            for atom in chi_atom:
                chi = apply(uniform, values[pos])
                if chi > 180.0: chi = chi - 360.0
                elif chi < -180.0: chi = chi + 360.0
                atom.tp_torsion = chi
                pos = pos + 1


def set_turn(p, res1, res2, generator, apply=apply):
    """Generate turn conformation for two sequentially successive
    residues

    Arguments

        o *res1*, *res2* - integers - indices of the two residues
          whose conformation is to be set to turn

        o *generator* - a random number generator

    Description

        The particular turn conformation chosen depends on the type
        of the two residues.  For the general case one of four
        possible turn types (Type I, Type II, Type I', Type II') is
        chosen and res1 is assigned values corresponding to the i+1
        position of the turn and res2 is assigned values corresponding
        to the i+2 position of a turn.  Only the Z-matrix description
        is modified by this function. Cartesian coordinates are not
        computed
    """    
    r1name = p.residue_names[res1]
    r2name = p.residue_names[res2]
    choice = generator.choice

    if r1name <> 'PRO' and r2name <> 'PRO':
        ms1, ms2 = choice(X_X_TURN_MS)
        f1, y1, w1 = MESO_VALUES[ms1]
        f2, y2, w2 = MESO_VALUES[ms2]
    elif r1name == 'PRO':
        if r2name == 'PRO':
            f1, y1, w1, f2, y2, w2 = choice(PRO_PRO_TURN)
        else:
            f1, y1, w1, f2, y2, w2 = choice(PRO_X_TURN)
    elif r2name == 'PRO':
        f1, y1, w1, f2, y2, w2 = choice(X_PRO_TURN)
        

    uniform = generator.uniform


    f1 = apply(uniform, f1)
    if f1 < -180.0:
        f1 = f1 + 360.0
    elif f1 > 180.0:
        f1 = f1 - 360.0

    y1 = apply(uniform, y1)
    if y1 < -180.0:
        y1 = y1 + 360.0
    elif y1 > 180.0:
        y1 = y1 - 360.0

    w1 = apply(uniform, w1)
    if w1 < -180.0:
        w1 = w1 + 360.0
    elif w1 > 180.0:
        w1 = w1 - 360.0

    f2 = apply(uniform, f2)
    if f2 < -180.0:
        f2 = f2 + 360.0
    elif f2 > 180.0:
        f2 = f2 - 360.0

    y2 = apply(uniform, y2)
    if y2 < -180.0:
        y2 = y2 + 360.0
    elif y2 > 180.0:
        y2 = y2 - 360.0

    w2 = apply(uniform, w2)
    if w2 < -180.0:
        w2 = w2 + 360.0
    elif w2 > 180.0:
        w2 = w2 - 360.0

    if p.residue_turn1_weights[res1] >= 1.0e-9:
        try:
            p.phi_atoms[res1].tp_torsion = f1
        except TypeError:
            pass
        else:
            p.phi_atoms[res1].sp_angle = uniform(111.0, 114.0)

        try:
            p.psi_atoms[res1].tp_torsion = y1
        except TypeError:
            pass

        try:
            p.omega_atoms[res1].tp_torsion = w1
        except TypeError:
            pass

        chi_atom = p.chi_atoms[res1]
        if chi_atom:
            values = choice(CHI_VALUES[r1name])
            pos = 0
            for atom in chi_atom:
                chi = apply(uniform, values[pos])
                if chi > 180.0: chi = chi - 360.0
                elif chi < -180.0: chi = chi + 360.0
                atom.tp_torsion = chi
                pos = pos + 1

    if p.residue_turn2_weights[res2] >= 1.0e-9:
        try:
            p.phi_atoms[res2].tp_torsion = f2
        except TypeError:
            pass
        else:
            p.phi_atoms[res2].sp_angle = uniform(111.0, 114.0)


        try:
            p.psi_atoms[res2].tp_torsion = y2
        except TypeError:
            pass


        try:
            p.omega_atoms[res2].tp_torsion = w2
        except TypeError:
            pass



        chi_atom = p.chi_atoms[res2]
        if chi_atom:
            values = generator.choice(CHI_VALUES[r2name])
            pos = 0
            for atom in chi_atom:
                chi = apply(uniform, values[pos])
                if chi > 180.0: chi = chi - 360.0
                elif chi < -180.0: chi = chi + 360.0
                atom.tp_torsion = chi
                pos = pos + 1



def set_mesostate_conformation(p, start, end, generator):
    """Set the conformation of a segment in accordance with the
    allowed mesostates for each residue in the segment.

    Arguments

        o *p* - instance of linusMol - the protein whose conformation is
        to be changed

        o *start* - integer - index of the first residue in the segment

        o *end* - integer - index of the last residue in the segment -
        *note* - only the conformation of residues start to end -1 are
        changed.

        o *generator* - object - a random number generator

    Description

        For each residue in the segment, a mesostate is chosen from
        the list of allowed mesostates for the residue.  Phi, psi and
        omega values in accordance with this mesostate are then
        generated.  Chi-values are chosen at random.  If no mesostates
        have been specified for a residue then its conformation remains
        unchanged.
    """    
    uniform = generator.uniform
    choice = generator.choice
    ms = p.mesostates
    rn = p.residue_names
    phi_atoms = p.phi_atoms
    psi_atoms = p.psi_atoms
    ome_atoms = p.omega_atoms
    chi_atoms = p.chi_atoms

    for i in range(start, end):
        if ms[i]:
            mesostate = choice(ms[i])
        else:
            continue
        values = MESO_VALUES[mesostate]
        f = apply(uniform, values[0])
        y = apply(uniform, values[1])
        w = apply(uniform, values[2])
        if f < -180.0:
            f = f + 360.0
        elif f > 180:
            f = f - 360.0
        if y < -180.0:
            y = y + 360.0
        elif y > 180:
            y = y - 360.0
        if w< -180.0:
            w = w + 360.0
        elif w > 180:
            w = w - 360.0

        try:
            phi_atoms[i].tp_torsion = f
        except TypeError:
            pass
        else:
            phi_atoms[i].sp_angle = uniform(108.0, 113.0)

        try:
            psi_atoms[i].tp_torsion = y
        except TypeError:
            pass

        try:
            ome_atoms[i].tp_torsion = w
        except TypeError:
            pass

        chi_atom = chi_atoms[i]
        if chi_atom:
            values = choice(CHI_VALUES[rn[i]])
            pos = 0
            for atom in chi_atom:
                chi = apply(uniform, values[pos])
                if chi > 180.0: chi = chi - 360.0
                elif chi < -180.0: chi = chi + 360.0
                atom.tp_torsion = chi
                pos = pos + 1

        
def set_residue_mesostate(p, start, end, ms, generator):

    choice = generator.choice
    uniform = generator.uniform

    rn = p.residue_names
    phi_atoms = p.phi_atoms
    psi_atoms = p.psi_atoms
    ome_atoms = p.omega_atoms
    chi_atoms = p.chi_atoms

    for i in range(start, end):
        mesostate = ms[i - start]
        values = MESO_VALUES[mesostate]
        f = apply(uniform, values[0])
        y = apply(uniform, values[1])
        w = apply(uniform, values[2])
        if f < -180.0:
            f = f + 360.0
        elif f > 180:
            f = f - 360.0
        if y < -180.0:
            y = y + 360.0
        elif y > 180:
            y = y - 360.0
        if w< -180.0:
            w = w + 360.0
        elif w > 180:
            w = w - 360.0

        try:
            phi_atoms[i].tp_torsion = f
        except TypeError:
            pass
        else:
            phi_atoms[i].sp_angle = uniform(111.0, 114.0)

        try:
            psi_atoms[i].tp_torsion = y
        except TypeError:
            pass

        try:
            ome_atoms[i].tp_torsion = w
        except TypeError:
            pass

        chi_atom = chi_atoms[i]
        if chi_atom:
            values = choice(CHI_VALUES[rn[i]])
            pos = 0
            for atom in chi_atom:
                chi = apply(uniform, values[pos])
                if chi > 180.0: chi = chi - 360.0
                elif chi < -180.0: chi = chi + 360.0
                atom.tp_torsion = chi
                pos = pos + 1
        
def choose_conformation(p, res_id, generator):
    """Choose a conformation given the index of a residue

    Arguments

        o *p* - instance of linusMol - the protein whose conformation is
        to be changed

        o *res_id* - integer - index of residue

    Description

        If mesostates have been specified for the residue then one of
        the allowed mesostates is chosen.  Otherwise one of helix,
        strand, turn1, turn2 and coil is chosen in accordance with the
        sampling weights of the residue.

    Result

        return a tuple of 2 elements, the first being a single character
        representation ('C' for mesostate choice, 'H' for helix, 'E' for
        strand, '1' for turn1, '2' for turn2 and 'C' for coil) of the
        move chosen and the second item being the function corresponding
        to the move chosen
    """    
        
    if hasattr(p, 'mesostates'):
        if p.mesostates[res_id]:
            return 'C', set_mesostate_conformation
    hw = p.residue_helix_weights[res_id]
    sw = hw + p.residue_strand_weights[res_id]
    t1w = sw + p.residue_turn1_weights[res_id]
    t2w = t1w + p.residue_turn2_weights[res_id]
    cw = t2w + p.residue_coil_weights[res_id]
    pw = cw + p.residue_PII_weights[res_id]
    r = generator.uniform(0.0, pw)
    if r <= hw:
        return 'H', set_helix
    elif r<=sw:
        return 'E', set_strand
    elif r<=t1w:
        return '1', set_turn
    elif r<=t2w:
        return '2', set_turn
    elif r<=cw:
        return 'C', set_coil
    elif r<=pw:
        return 'P', set_PII
    else:
        return None, None


try:
    import psyco
except ImportError:
    pass
else:
    psyco.bind(set_helix)
    psyco.bind(set_strand)
    psyco.bind(set_turn)
    psyco.bind(set_coil)
    psyco.bind(set_PII)
    psyco.bind(set_mesostate_conformation)
    psyco.bind(choose_conformation)
    psyco.bind(set_residue_mesostate) 
    psyco.bind(set_super_helix)
