"""Module to assign conformational mesostates and secondary structure
to a molecule.

Functions in this module

    o *rc_gen* - assign mesostates to a molecule

    o *rc_ss* - assign secondary structure given mesostates
"""
    
RC_CODES = (
    ('A', 'B', 'C', 'D', 'E', 'F'),
    ('G', 'H', 'I', 'J', 'K', 'L'),
    ('M', 'N', 'O', 'P', 'Q', 'R'),
    ('S', 'T', 'U', 'V', 'W', 'X'),
    ('m', 'r', 'q', 'p', 'o', 'n'),
    ('g', 'l', 'k', 'j', 'i', 'h')
    )

from string import join

def rc_gen(mol, codes=None, RC_CODES=RC_CODES):
    """Assign mesostate conformational codes to a molecule

    Arguments

        o *mol* - instance of LinusMol

        o *codes* - list - this is an optional argument and if supplied
        will contain the mesostates for each residue.  The minimun length
        of the list is the number of residues in the molecule

    Result

        A list containing the mesostate assignments for each residue
        in the molecule
    """
    
    if codes is None:
        codes = ['?']*mol.num_residues
        
    phi_atoms = mol.phi_atoms
    psi_atoms = mol.psi_atoms
    ome_atoms = mol.omega_atoms

    for i in range(1, mol.num_residues):
        phi = phi_atoms[i].tp_torsion
        psi = psi_atoms[i].tp_torsion + 180.0
        if psi > 180: psi = psi - 360.0
        cis = abs(ome_atoms[i].tp_torsion) <= 90.0

        if cis:
            codes[i] = '*'
        else:
            if phi<-150.0 or phi>150.0:
                row = RC_CODES[0]
            elif phi <= -90.0:
                row = RC_CODES[1]
            elif phi <= -30.0:
                row = RC_CODES[2]
            elif phi <=  30.0:
                row = RC_CODES[3]
            elif phi <=  90.0:
                row = RC_CODES[4]
            else:
                row = RC_CODES[5]
                
            if psi < -150.0 or psi > 150.0:
                codes[i] = row[0]
            elif psi < -90.0:
                codes[i] = row[1]
            elif psi < -30.0:
                codes[i] = row[2]
            elif psi < 30.0:
                codes[i] = row[3]
            elif psi < 90.0:
                codes[i] = row[4]
            else:
                codes[i] = row[5]
    return join(codes, '')

import re

helix_begin = re.compile('[OP]{5}')
helix_end = re.compile('[OP](?![OP])')
strand_begin = re.compile('[LG]{3}')
strand_end = re.compile('[LG](?![LG])')

def find_helices(rc):
    start = 0
    finish = len(rc)
    matches = []
    while start < finish:
        res = helix_begin.search(rc,start)
        if res is None: break
        begin = res.regs[0][0]
        res = helix_end.search(rc, res.regs[0][1]-1)
        end = res.regs[0][1]
        matches.append((begin, end))
        start = end + 1
    return matches

def find_strands(rc):
    start = 0
    finish = len(rc)
    matches = []
    while start < finish:
        res = strand_begin.search(rc,start)
        if res is None: break
        begin = res.regs[0][0]
        res = strand_end.search(rc, res.regs[0][1]-1)
        end = res.regs[0][1]
        matches.append((begin, end))
        start = end + 1
    return matches

turns = ('OO', 'OP', 'OJ', 'PO', 'PP', 'PJ', 'JO', 'JP', 'JJ',
         'Mo', 'Mp', 'Mj', 'Ro', 'Rp', 'Rj',
         'oo', 'op', 'oj', 'po', 'pp', 'pj', 'jo', 'jp', 'jj',
         'mO', 'mP', 'mJ', 'rO', 'rP', 'rJ')
trndic = {}
for i in range(len(turns)):
     trndic[turns[i]] = 1

PIIdic = {'M':1, 'R':1}

def rc_ss(rc, sst=None):
    """Assign secondary structure from mesostates

    Arguments

        o *rc* - string or list - mesostates of each residue in a
        molecule

        o *sst* - list - this is an optional argument.  If supplied
        the secondary structure assignment will be written to this
        list.  The list must have a minimum length equal to the
        length of *rc*

    Result

        The secondary structure of each residue as a list
    """    
    is_turn = trndic.has_key
    is_PII = PIIdic.has_key
    if sst is None:
        sst = ['C']*len(rc)
    else:
        for i in range(len(sst)): sst[i] = 'C'

    # assign PII
    for i in range(1, len(rc)):
        ss = rc[i]
        if is_PII(ss):
            sst[i] = 'P'

    # assign turns
    for i in range(1, len(rc)-1):
        i1 = i + 1
        ss = rc[i]+rc[i1]
        if is_turn(ss):
            sst[i] = '1'
            sst[i1] = '2'

    # assign helices
    helices = find_helices(rc)
    if helices:
        for i, j in helices:
            for k in range(i, j): sst[k] = 'H'

    # assign strands
    strands = find_strands(rc)
    if strands:
        for i, j in strands:
            for k in range(i, j): sst[k] = 'E'
    return join(sst, '')

def mol_ss(mol):
    return rc_ss(rc_gen(mol))

try:
    import psyco
except ImportError:
    HAVE_PSYCO = 0
else:
    HAVE_PSYCO = 1

try:
    from linusc import rc_gen
except ImportError:
    if HAVE_PSYCO:
        psyco.bind(rc_gen)

if HAVE_PSYCO:
   psyco.bind(rc_ss)
   psyco.bind(find_helices)
   psyco.bind(find_strands)
   psyco.bind(mol_ss)
