import math

try:
    import Numeric
except ImportError:
    HAVE_NUMPY = 0
else:
    HAVE_NUMPY = 1

_RAD_TO_DEG = 180.0/math.pi
_DEG_TO_RAD = math.pi/180.0

def is_atom(o):
    return hasattr(o, '_is_an_atom3d')

class Atom3d:
    _is_an_atom3d = 1

    def __init__(self, x, y, z, cnf={}):
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)
        if cnf:
            for key, value in cnf.items():
                self.__dict__[key] = value

    def __repr__(self):
        return "Atom3d(%s, %s, %s)" % (self.x, self.y, self.z)
    __str__ = __repr__

    def coords(self):
        if HAVE_NUMPY:
            return Numeric.array((self.x, self.y, self.z))
        else:
            return [self.x, self.y, self.z]

    def set_coords(self, x, y=None, z=None):
        if y is None:
            self.x, self.y, self.z = map(float, x)
        else:
            self.x, self.y, self.z = float(x), float(y), float(z)

    def xcoord(self, v=None):
        if v is None:
            return self.x
        else:
            self.x = float(v)

    def ycoord(self, v=None):
        if v is None:
            return self.y
        else:
            self.y = float(v)

    def zcoord(self, v=None):
        if v is None:
            return self.z
        else:
            self.z = float(v)

    def clone(self):
        new = Atom3d(0., 0., 0.)
        new.__dict__.update(self.__dict__)
        return new

    __copy__ = clone

    def __deepcopy__(self):
        cnf = copy.deepcopy(self.__dict__)
        return self.__class__(0, 0, 0, cnf)

    def distance(self, other, sqrt=math.sqrt):
        if not hasattr(other, '_is_an_atom3d'):
            raise TypeError, 'distance argument must be Atom3d type'
        
        return sqrt( (self.x - other.x)**2 +
                     (self.y - other.y)**2 +
                     (self.z - other.z)**2 )

    def sqr_distance(self, other):
        if not hasattr(other, '_is_an_atom3d'):
            raise TypeError, 'sqr_distance argument must be Atom3d type'

        return (self.x - other.x)**2 + \
               (self.y - other.y)**2 + \
               (self.z - other.z)**2

    def angle(self, a, b, sqrt=math.sqrt, acos=math.acos):
        if not hasattr(a, '_is_an_atom3d') or \
           not hasattr(b, '_is_an_atom3d'):
            raise TypeError, 'angle arguments must be Atom3d type'

        x1, y1, z1 = self.x - a.x, self.y - a.y, self.z - a.z
        x2, y2, z2 = b.x - a.x, b.y - a.y, b.z - a.z
        v11 = x1**2 + y1**2 + z1**2
        v22 = x2**2 + y2**2 + z2**2
        
        if (v11 == 0.0 or v22 == 0.0):
            raise ValueError, 'Null vector in angle'
        
        v12 = x1*x2 + y1*y2 + z1*z2
        
        ang = v12/sqrt(v11*v22)
        
        if ang >= 1.0:
            return 0.0
        elif ang <= -1.0:
            return 180.0
        else:
            return acos(ang) * _RAD_TO_DEG

    def torsion(self, a, b, c, sqrt=math.sqrt, acos=math.acos):
        
        if not hasattr(a, '_is_an_atom3d') or \
           not hasattr(b, '_is_an_atom3d') or \
           not hasattr(c, '_is_an_atom3d'):
            raise TypeError, 'torsion arguments must be Atom3d type'
        
        v12x, v12y, v12z = self.x - a.x, self.y - a.y, self.z - a.z
        v32x, v32y, v32z = b.x - a.x, b.y - a.y, b.z - a.z
        v43x, v43y, v43z = c.x - b.x, c.y - b.y, c.z - b.z

        vn13x = v12y*v32z - v12z*v32y
        vn13y = v12z*v32x - v12x*v32z
        vn13z = v12x*v32y - v12y*v32x

        vn24x = v32z*v43y - v32y*v43z
        vn24y = v32x*v43z - v32z*v43x
        vn24z = v32y*v43x - v32x*v43y

        v12 = vn13x*vn24x + vn13y*vn24y + vn13z*vn24z
        v11 = vn13x**2 + vn13y**2 + vn13z**2
        v22 = vn24x**2 + vn24y**2 + vn24z**2

        ang = v12/sqrt(v11*v22)
        if ang >= 1.0:
            return 0.0
        elif ang <= -1.0:
            return -180.0
        else:
            ang = acos(ang) * _RAD_TO_DEG
            
        vtmp = vn13x * (vn24y*v32z - vn24z*v32y) + \
               vn13y * (vn24z*v32x - vn24x*v32z) + \
               vn13z * (vn24x*v32y - vn24y*v32x) < 0.0
        if vtmp:
            return -ang
        else:
            return ang

    def translate(self, a, b=None, c=None):
        if b is None:
            if hasattr(a, '_is_an_atom3d'):
                a, b, c = a.x, a.y, a.z
            else:
                a, b, c = map(float, a)
        else:
            a, b, c = map(float, (a, b, c))
        self.x = self.x + a
        self.y = self.y + b
        self.z = self.z + c

    def rotatex(self, theta):
        trad = theta * _DEG_TO_RAD
        sint, cost = math.sin(trad), math.cos(trad)
        y, z = self.y, self.z
        self.y = y*cost - z*sint
        self.z = y*sint + z*cost
        
    def rotatey(self, theta):
        trad = theta * _DEG_TO_RAD
        sint, cost = math.sin(trad), math.cos(trad)
        x, z = self.x, self.z
        self.x = x*cost + z*sint
        self.z = z*cost - x*sint
 
    def rotatez(self, theta):
        trad = theta * _DEG_TO_RAD
        sint, cost = math.sin(trad), math.cos(trad)
        x, y = self.x, self.y
        self.x = x*cost - y*sint
        self.y = x*sint + y*cost
#
# functions that operate on a collection of atoms 
#

if HAVE_NUMPY:
    def coords_as_array(list):
        """returns the coordinates of a collection of atoms as a Numeric array"""
        assert(len(list) > 0)
        a = Numeric.zeros((len(list), 3), 'd')
        pos = 0
        for atom in list:
            a[pos] = atom.coords()
            pos = pos + 1
        return a

def coords_as_list(list):
    """returns the coordinates of a collection of atoms as a list of lists"""
    assert(len(list)>0)
    a = [None]*len(list)
    pos = 0
    for atom in list:
        a[pos] = atom.coords()
        pos = pos + 1
    return a

def geometric_center(list):
    """returns the geometric center of a collection of atoms"""
    assert(len(list) > 0)
    x = y = z = 0.0
    for atom in list:
        x = x + atom.x
        y = y + atom.y
        z = z + atom.z
    n = 1.0/len(list)
    x = x * n; y = y * n; z = z * n;
    return [x, y, z]

def radius_of_gyration(list):
    """returns the radius of gyration of a collection of atoms"""
    cx, cy, cz = geometric_center(list)
    rg = 0.0
    for atom in list:
        rg = rg + (atom.x - cx)**2 + (atom.y - cy)**2 + (atom.z - cz)**2
    return math.sqrt(rg/len(list))

def fromint(v1, dis, v2, a, v3, t, sqrt=math.sqrt, sin=math.sin, cos=math.cos):

    ang = a * _DEG_TO_RAD
    sina = sin(ang)
    cosa = cos(ang)

    tors = t * _DEG_TO_RAD 
    sint = sina * sin(tors)
    cost = sina * cos(tors)

    x1, y1, z1 = v1.coords()
    x2, y2, z2 = v2.coords()
    x3, y3, z3 = v3.coords()
    
    u1x = x2 - x3
    u1y = y2 - y3
    u1z = z2 - z3
    d = 1.0 / sqrt(u1x*u1x + u1y*u1y + u1z*u1z)
    u1x = u1x*d
    u1y = u1y*d
    u1z = u1z*d

    u2x = x1 - x2
    u2y = y1 - y2
    u2z = z1 - z2;
    d = 1.0 / sqrt(u2x*u2x + u2y*u2y + u2z*u2z)
    u2x = u2x*d
    u2y = u2y*d
    u2z = u2z*d

    cosine = u1x*u2x + u1y*u2y + u1z*u2z

    if (abs(cosine) < 1.0):
        sine = 1.0/sqrt(1.0 - cosine*cosine)
    else:
        sine = 1.0/sqrt(cosine*cosine - 1.0)
            
    u3x = sine * (u1y*u2z - u1z*u2y)
    u3y = sine * (u1z*u2x - u1x*u2z)
    u3z = sine * (u1x*u2y - u1y*u2x)

    u4x = cost * (u3y*u2z - u3z*u2y)
    u4y = cost * (u3z*u2x - u3x*u2z)
    u4z = cost * (u3x*u2y - u3y*u2x)

    return Atom3d(x1 + dis*(-u2x*cosa + u4x + u3x*sint),
                  y1 + dis*(-u2y*cosa + u4y + u3y*sint),
                  z1 + dis*(-u2z*cosa + u4z + u3z*sint))



try:
    from _atom3d import *
except ImportError:
    pass
