import math
import types
from TypedList import TypedList
from exceptions import Exception
import Atom3d

HAVE_NUMPY = Atom3d.HAVE_NUMPY

if HAVE_NUMPY:
    import Numeric
else:
    class NoNUMPY_Error(Exceptions):
        def __init__(self):
            Exception.__init__('Function/Method requires NumPy Module')

is_atom = Atom3d.is_atom
    
def is_residue(o):
    return hasattr(o, '_is_a_residue')

def is_chain(o):
    return hasattr(o, '_is_a_chain')

def is_mol(o):
    return hasattr(o, '_is_a_mol')


class Residue(TypedList):
    _is_a_residue = 1

    def __init__(self, name='', atoms=None, **kw):
        TypedList.__init__(self, is_atom, atoms)
        self.name = name
        for key, value in kw.items():
            setattr(self, key, value)        
        
    def num_atoms(self):
        """returns the number of atoms in residue"""
        return len(self.elements)

    def has_atom(self, name):
        """returns true if residue has an atom named 'name'"""
        for atom in self.elements:
            if atom.name == name:
                return 1
        return 0

    def atoms(self):
        return self.elements

    def atoms_with_name(self, *names):
        """returns atoms in residue with specified names"""
        ret = []
        Append = ret.append
        for name in names:
            for atom in self.elements:
                if atom.name == name:
                    Append(atom)
                    break
        return ret

    def delete_atoms_with_name(self, *names):
        """delete atoms in residue with specified names"""
        els = self.elements
        for i in self.reverse_indices():
            atom = els[i]
            if atom.name in names:
                del els[i]

    def atoms_not_with_name(self, *names):
        """returns atoms in residue excluding specified names"""
        ret = []
        for atom in self.elements:
            if not atom.name in names:
                ret.append(atom)
        return ret

    def atom_coordinates(self, *names):
        """returns coordinates of named atoms.
        If names is omitted all atom coordinates are returned."""

        if len(names)==0:
            atoms = self.elements
        else:
            atoms = apply(self.atoms_with_name, names)

        na = len(atoms)
        if na == 0: return

        if HAVE_NUMPY:
            a = Numeric.zeros((na, 3), 'd')
        else:
            a = [None]*na

        pos = 0
        for atom in atoms:
            a[pos] = atom.coords()
            pos = pos + 1
        return a

    def geometric_center(self):
        return Atom3d.geometric_center(self.elements)

    def extents(self):
        """returns a list of the minimum and maximum, x, y and z coordinates"""
        coords = self.atom_coordinates()
        ret = _extents(coords)
        del coords
        return ret

    def assign_radii(self):
        raise AttributeError, 'Should be defined in specialized class'

    def translate(self, dx, dy=None, dz=None):
        if dy is None:
            dx, dy, dz = dx
        for atom in self.elements:
            atom.translate(dx, dy, dz)

    def rotatex(self, theta):
        for atom in self.elements:
            atom.rotatex(theta)

    def rotatey(self, theta):
        for atom in self.elements:
            atom.rotatey(theta)

    def rotatez(self, theta):
        for atom in self.elements:
            atom.rotatez(theta)

    def type(self):
        return 'residue'

class Chain(TypedList):
    _is_a_chain = 1

    def __init__(self, name='', residues=None):
        self.name = name
        TypedList.__init__(self, is_residue, residues)

    def num_residues(self): return len(self)

    def num_atoms(self):
        na = 0
        for res in self.elements:
            na = na + len(res.elements)
        return na

    def atoms(self):
        ret = []
        Append = ret.append
        for res in self.elements:
            for atom in res.elements:
                Append(atom)
        return ret

    def residues_with_name(self, *names):
        """returns named residues as a python list"""
        if len(names) == 0:
            return
        l = []
        for res in self.elements:
            if res.name in names:
                l.append(res)
        return l

    def delete_residues_with_name(self, *names):
        """delete named residues from Chain"""
        if len(names) == 0:
            return
        els = self.elements
        for i in self.reverse_indices():
            if els[i].name in names:
                del els[i]

    def residues_not_with_name(self, *names):
        """returns residues excluding specified names as a python list"""
        ret = []
        for res in self.elements:
            if not res.name in names:
                ret.append(res)
        return ret
        

    def atoms_with_name(self, *names):
        ret = []
        Append = ret.append
        for res in self.elements:
            for name in names:
                for atom in res.elements:
                    if atom.name == name:
                        Append(atom)
                        break
        return ret
            
    def delete_atoms_with_name(self, *names):
        """delete atoms in residue with specified names"""
        for res in self.elements:
            apply(res.delete_atoms_with_name, names)

    def atoms_not_with_name(self, *names):
        """returns atoms in residue excluding specified names"""
        ret = []
        for res in self.elements:
            ret[len(ret):] = apply(res.atoms_not_with_name, names)
        return ret

    def atom_coordinates(self, *names):
        """returns coordinates of named atoms. if names is None
        all atom coordinates are returned."""

        coords = []
        if len(names) > 0:
            for res in self.elements:
                for atom in res.elements:
                    if atom.name in names:
                        coords.append(atom.coords())
        else:
            atoms = apply(self.atoms_with_name, names)
            coords = map(lambda a:a.coords(), atoms)
        if HAVE_NUMPY:
            return Numeric.array(coords)
        else:
            return coords

    def atoms(self):
        atoms = []
        for res in self.elements:
            for atom in res: atoms.append(atom)
        return atoms

    def geometric_center(self):
        """returns the geometric center of the chain"""
        atoms = self.atoms()
        gc = _atom3d.geometric_center(atoms)
        del atoms
        return gc

    def radius_of_gyration(self):
        """returns the radius of gyration of the chain."""
        atoms = self.atoms()
        rg = Atom3d.radius_of_gyration(atoms)
        del atoms
        return rg

    def assign_radii(self):
        for res in self:
            res.assign_radii()

    def delete_hetero(self):
        for i in xrange(self.num_residues()-1, -1, -1):
            if hasattr(self[i], 'het'):
                del self[i]

    def delete_waters(self):
        for i in xrange(self.num_residues()-1, -1, -1):
            if self[i].name == 'HOH':
                del self[i]

    def extents(self):
        """returns the minimum and maximum x, y and z coords of chain"""
        coords = self.atom_coordinates()
        ret = _extents(coords)
        del coords
        return ret

    def translate(self, dx, dy=None, dz=None):
        for res in self.elements:
            res.translate(dx, dy, dz)
    
    def rotatex(self, theta):
        for res in self.elements:
            res.rotatex(theta)

    def rotatey(self, theta):
        for res in self.elements:
            res.rotatey(theta)

    def rotatez(self, theta):
        for res in self.elements:
            res.rotatez(theta)

    def type(self):
        return 'Chain'

class Mol(TypedList):
    _is_a_mol = 1

    def __init__(self, name='', chains=None):
        self.name = name
        TypedList.__init__(self, is_chain, chains)

    def num_chain(self): return len(self)

    def num_res(self):
        nr = 0
        for chain in self: nr = nr + len(chain)
        return nr

    def num_atoms(self):
        na = 0
        for chain in self: na = na + chain.num_atoms()
        return na

    def atoms(self):
        ret = []
        Append = ret.append

        for chain in self.elements:
            for res in chain.elements:
                for atom in res.elements:
                    Append(atom)
        return ret

    def chains_with_name(self, *names):
        chains = []
        for chain in self.elements:
            if chain.name in names: chains.append(chain)
        return chains

    def residues_with_name(self, *names):
        residues = []
        for chain in self.elements:
            for res in chain.elements:
                if res.name in names: residues.append(res)
        return residues


    def atoms_with_name(self, *names):
        ret = []
        Append = ret.append
        for chain in self.elements:
            for res in chain.elements:
                for name in names:
                    for atom in res.elements:
                        if atom.name == name:
                            Append(atom)
                            break
        return ret

    def delete_atoms_with_name(self, *names):
        """delete atoms in assembly with specified names"""
        for chain in self.elements:
            apply(chain.delete_atoms_with_name, names)

    def atoms_not_with_name(self, *names):
        """returns atoms in assembly excluding specified names"""
        ret = []
        for chain in self.elements:
            ret[len(ret):] = apply(chain.atoms_not_with_name, names)
        return ret


    def atom_coordinates(self, *names):

        coords = []
        if len(names) == 0:
            for chain in self.elements:
                for res in chain.elements:
                    for atom in res.elements:
                        coords.append(atom.coords())
        else:
            atoms = apply(self.atoms_with_name, names)
            coords = map(lambda a:a.coords(), atoms)
            del atoms
        if HAVE_NUMPY:
            return Numeric.array(coords)
        else:
            return coords

    def geometric_center(self):
        atoms = self.atoms()
        gc = Atom3d.geometric_center(atoms)
        del atoms
        return gc

    def center(self):
        dx, dy, dz = self.geometric_center()
        dx = -dx
        dy = -dy
        dz = -dz
        for chain in self.elements:
            for res in chain.elements:
                for atom in res.elements:
                    atom.translate(dx, dy, dz)

    def radius_of_gration(self):
        atoms = self.atoms()
        rg = Atom3d.radius_of_gyration(atoms)
        del atoms
        return rg

    def assign_radii(self):
        for chain in self:
            chain.assign_radii()

    def delete_water(self):
        for chain in self:
            chain.delete_water()

    def delete_hetero(self):
        for i in self.reverse_indices():
            self[i].delete_hetero()
            if len(self[i]) == 0:
                del self[i]

    def translate(self, dx, dy=None, dz=None):
        for chain in self:
            chain.translate(dx, dy, dz)

    def rotatex(self, theta):
        for chain in self:
            chain.rotatex(theta)

    def rotatey(self, theta):
        for chain in self:
            chain.rotatey(theta)

    def rotatez(self, theta):
        for chain in self:
            chain.rotatez(theta)
    
    def extents(self):
        return _extents(self.atom_coordinates())

                            
def _extents(c):
    if HAVE_NUMPY:
        xmin, ymin, zmin = Numeric.minimum.reduce(c)
        xmax, ymax, zmax = Numeric.maximum.reduce(c)
        ret = Numeric.array([xmin, xmax, ymin, ymax, zmin, zmax])
    else:
        ret = []
        for i in (0, 1, 2):
            t = map(lambda a:a[i], c)
            ret.append((min(t), max(t)))
    return ret
        
          

        
