import string
import gzip
from types import FileType
import mol
import Atom3d
import protein, pro_dic
import rna, rna_dic
import dna, dna_dic


def _atof(s, atof=string.atof):
    try:
        return atof(s)
    except:
        return None

def _atoi(s, atoi=string.atoi):
    try:
        return atoi(s)
    except:
        return None

def get_sequences(file):
    if file[-3:] == '.gz':
        ff = gzip.GzipFile(file)
    else:
        ff = open(file, 'r')
    
    # read until we find a line with SEQRES card
    line = ff.readline()
    while line:
        if line[:6] == 'SEQRES':
            break
        else:
            line = ff.readline()

    if not line: # no SEQRES records
        return

    sequences = {}
    chain = line[11]
    sequences[chain] = []
    
    while line[:6] == 'SEQRES':
        chain = line[11]
        residues = string.split(line[19:71])
        if not sequences.has_key(chain):
            sequences[chain] = []
        sequences[chain].extend(residues)
        line = ff.readline()

    return sequences
     

def unpack_pdb_line(line, ATOF=_atof, ATOI=_atoi, STRIP=string.strip):
    return (ATOI(line[6:11]),
           STRIP(line[12:16]),
           line[16],
           STRIP(line[17:21]),
           line[21],
           STRIP(line[22:27]),
           ATOF(line[30:38]),
           ATOF(line[38:46]),
           ATOF(line[46:54]),
           ATOF(line[54:60]),
           ATOF(line[60:66]))


def atom_build(t, atom=Atom3d.Atom3d):
    atm = atom(t[6], t[7], t[8]); atm.occ = t[9]; atm.bf = t[10]; atm.name = t[1]; atm.idx = t[0];
    return atm

def _type_mol(mol):
    if not len(mol):
        return mol
    for i in range(len(mol)):
        chain = mol[i]
        if is_protein(chain):
            Chain = protein.Protein
            Residue = protein.AminoAcid
        elif is_rna(chain):
            Chain = rna.RNA
            Residue = rna.RNABase
        elif is_dna(chain):
            Chain = dna.DNA
            Residue = dna.DNABase
        else:
            continue

        new_chain = Chain(chain.name)
        Append = new_chain.elements.append
        for res in chain.elements:
            new_res = Residue(res.name, idx=res.idx)
            new_res.elements = res.elements[:]
            Append(new_res)
            if hasattr(res, 'het'):
                new_res.het = 1
        mol[i] = new_chain
        del chain
    return mol

def is_protein(chain):
    for res in chain.elements:
        if res.name in pro_dic.RESIDUES:
            return 1
        elif hasattr(res, 'het'):
            return 0
    return 0


def is_dna(chain):
    for res in chain.elements:
        if res.name in dna_dic.RESIDUES:
            if not res.has_atom('O2*'): return 1
        elif hasattr(res, 'het'):
            return 0
    return 0

def is_rna(chain):
    for res in chain.elements:
        if res.name in rna_dic.RESIDUES:
            if res.has_atom('O2*'): return 1
        elif hasattr(res, 'het'):
            return 0
    return 0
    
def read_pdb(f, as_protein=0, as_rna=0, as_dna=0, unpack=unpack_pdb_line,
              atom_build=atom_build):

    if type(f) is FileType:
        inp = f
        close = 0
    else:
        close = 1
        if f[-3:] == '.gz':
            inp = gzip.GzipFile(f)
        else:
            inp = open(f, 'r')

    try:
        sequences = get_sequences(inp.name)
    except AttributeError:
        sequences = get_sequences(inp.filename)

    Mol = mol.Mol
    Chain = mol.Chain
    Residue = mol.Residue

    if as_protein:
        Chain = protein.Protein
        Residue = protein.AminoAcid
    elif as_rna:
        Chain = rna.RNA
        Residue = rna.RNABase
    elif as_dna:
        Chain = dna.DNA
        Residue = dna.DNABase

    new_mol = Mol('')

    nextline = inp.readline
    while 1:
        line = nextline()
        if not line:
            break
        if line[:6] == 'COMPND':
            if string.find(line, 'MOL_ID') <> -1: continue
            new_mol.name = new_mol.name + string.strip(line[10:72]) + ' '
        if line[:4] == 'ATOM' or line[:4] == 'HETA' or line[:5] == 'MODEL':
            break
    new_mol.name = string.strip(new_mol.name)    
    data = inp.readlines()
    if line[:4] in ('ATOM', 'HETA', 'MODE'):
        data.insert(0, line)

    old_chain = None
    old_res = None

    for line in data:
        key = line[:4]
        if key == 'ATOM' or key == 'HETA':
            t = unpack(line)
            if t[4] <> old_chain:
                old_chain = t[4]
                new_chain = Chain(old_chain)
                AppendResidue = new_chain.elements.append
                new_mol.append(new_chain)
                old_res = None
            if t[5] <> old_res:
                old_res = t[5]
                new_res = Residue(t[3], idx=old_res)
                if key == 'HETA' and not t[3] in sequences.get(t[4], (' ',)):
                    new_res.het = 1
                AppendAtom = new_res.elements.append
                AppendResidue(new_res)
                seen = []
            if not t[1] in seen: 
                AppendAtom(atom_build(t))
                seen.append(t[1])
        elif key == 'MODE':
            if len(new_mol) > 0:
                break
            new_chain = Chain(string.split(line)[1])
            AppendResidue = new_chain.elements.append
            new_mol.append(new_chain)
            old_chain = ' '
    del data
    if close:
        inp.close()

    if as_protein or as_dna or as_rna:
        return new_mol
    else:
        return _type_mol(new_mol)

def _read_chain(f, as_protein=0, as_rna=0, as_dna=0, unpack=unpack_pdb_line,
                atom_build = atom_build):

    if as_protein:
        Chain = protein.Protein
        Residue = protein.AminoAcid
    elif as_rna:
        Chain = rna.RNA
        Residue = rna.RNABase
    elif as_dna:
        Chain = dna.DNA
        Residue = dna.DNABase
    else:
        Chain = mol.Chain
        Residue = mol.Residue

    nextline = f.readline

    new_chain = Chain('')
    AppendResidue = new_chain.elements.append
    old_chain = ' '
    old_res = None

    while 1:
        line = nextline()
        if not line:
            break
        key = line[:4]
        if key == 'ATOM' or key == 'HETA':
            break
        elif key == 'MODE':
            new_chain.name = string.split(line)[1]
            line = None

    if line:
        t = unpack(line)
        new_res = Residue(t[3], idx=t[5])
        AppendResidue(new_res)
        old_res = t[5]
        AppendAtom = new_res.elements.append
        AppendAtom(atom_build(t))
        ResidueHasAtom = new_res.atoms_with_name

    while 1:
        line = nextline()
        if not line:
            break
        key = line[:4]
        if key == 'ATOM' or key == 'HETA':
            t = unpack(line)
            if t[5] <> old_res:
                old_res = t[5]
                new_res = Residue(old_res, idx=t[5])
                AppendResidue(new_res)
                AppendAtom = new_res.elements.append
                ResidueHasAtom = new_res.atoms_with_name
            if not ResidueHasAtom(t[1]):
                AppendAtom(atom_build(t))

        elif key[:3] in ('TER', 'END'):
            break

    if as_protein or as_dna or as_rna:
        return new_chain
    else:
        return _type_mol([new_chain])[0]
            
class PDBFile:

    def __init__(self, filename):
        self.filename = filename
        self._openfile()

    def _openfile(self):
        if self.filename[-3:] == '.gz':
            self.file = gzip.GzipFile(self.filename)
        else:
            self.file = open(self.filename)

    def read(self, as_protein=0, as_rna=0, as_dna=0):
        return read_pdb(self.file, as_protein, as_rna, as_dna)

    def read_chain(self, as_protein=0, as_rna=0, as_dna=0):
        return _read_chain(self.file, as_protein, as_rna, as_dna)

if __name__ == "__main__":
    import time
    t1 = time.clock()
    mol = read_pdb('/usr2/pdb/pdb2hhb.ent.gz')
    t2 = time.clock()
    print t2 - t1
