/****************************************************************
 *This module does various hydrogen bond calculation routines.  *
 *The major ones are:                                           * 
 *1) intra_peptide_hbonds (bb-bb hbonds)                        *
 *2) water_hbonds (bb-water hbonds)                             * 
 *3) bb_sc_sat    (bb-sc hbonds)                                *
 *It is used for hydrogen bond satisfaction in LINUS            *
 ****************************************************************/

/*Last update: 8/28/07*/


#include "Python.h"
#include <stdio.h>
#include "atom3d.h"
#include <math.h>
#include <string.h>
#include <stdio.h>

/*Global Variables*/
#define DEGREES_TO_RADIANS 0.017453292519943295e0

/*Macros*/
#define ATOM3D_DATA(ob)  (Atom3d *)(((PyArrayObject *)(ob))->data)
#define GET_ATOM(ob,i) (Atom3d *)(PyList_GET_ITEM((ob),(i)))
#define GET_INT(ob,i) (int )  (PyInt_AS_LONG(PyList_GET_ITEM((ob),(i))))
#define GET_FLT(ob,i) (double)(PyFloat_AS_DOUBLE(PyList_GET_ITEM((ob),(i))))
#define GET_DICT_DBL(ob,i,v) PyArg_Parse(PyDict_GetItemString((ob),i),"d",&v)
#define GET_STR(ob,i) (char *)PyString_AS_STRING((PyList_GET_ITEM((ob),(i))))


/***********************Begin functions for general use***********************/
double max(double a,double b)
{
  if (a>b)  return a;
  return b;
}

double min(double a,double b)
{
  if (a<b)  return a;
  return b;
}

void set_list(int list[],int size,int value)
{
  int i;

  for (i=0; i<size; i++){
    list[i] = value;
  }
  return;
}

void make_atom(Atom3d *fp, float d, Atom3d *sp, float a, Atom3d *tp, 
               float t, Atom3d *self)
{

  self->fp_distance   = d;
  self->sp_angle      = a;
  self->tp_torsion    = t;
  self->first_parent  = fp;
  self->second_parent = sp;
  self->third_parent  = tp;

  Atom3d_IntToCart(self);

  return;
}

void atoms_with_name(char *atmname, PyObject* atoms, Atom3d* specific_atoms[])
{
  int j = 0;
  int i;
  int size = PyList_GET_SIZE(atoms);

  for (i=0; i<size; i++){
    Atom3d *atom = (Atom3d *)PyList_GET_ITEM(atoms,i);
    if (strcmp(atmname,atom->name) == 0) {
      specific_atoms[j] = atom;
      j = j+1;
    }
  }
  return;
}

int unsatisfied(value)
{
  return (value == -1);
}

int sc_satisfied(long value, PyObject *first_atom_indices)
{

  int i;
  int size = PyList_GET_SIZE(first_atom_indices);

  for (i=0;i<size;i++){
    if (value == GET_INT(first_atom_indices,i)+1) return 1;
  }

  return 0;
}
/****************************End functions for general use********************/


/*********************Begin Python Memory Handling Functions******************/
void Atom3d_dealloc(Atom3d *self)
{
  Py_XDECREF(self->first_parent);
  Py_XDECREF(self->second_parent);
  Py_XDECREF(self->third_parent);
  PyMem_DEL(self);
}

void Atom3d_parents_INCREF(Atom3d *fp, Atom3d *sp, Atom3d *tp)
{
  Py_INCREF(fp);
  Py_INCREF(sp);
  Py_INCREF(tp);
}

void Atom3d_parents_XDECREF(Atom3d *fp, Atom3d *sp, Atom3d *tp)
{
  Py_XDECREF(fp);
  Py_XDECREF(sp);
  Py_XDECREF(tp);
}
/***********************End Python Memory Handling Functions******************/


/*************************Begin backbone hbond utilities**********************/

/*This function determines if bb donors and acceptors are hbonded using
  geometric criteria: distance, C-O-N angle, C-N-O angle, and C-O-N-C torsion*/
    
int hbonded(Atom3d *don, Atom3d *acc, Atom3d *dfp, Atom3d *afp, double dist,
	    double theta1, double theta2, double torsion, int radius)
{

  /*Hydrogen bonds cannot be satisfied outside of the radius, through the 
    peptide bond, or in the same residue*/
  if (!((abs(acc->resnum-don->resnum) <= radius) &&     
	(acc->resnum-don->resnum != -1) && (acc->resnum-don->resnum != 0))){
    /*printf("Donor %d, Acceptor %d, rejected for outside radius\n",
      don->resnum,acc->resnum);*/
    return 0;
  }
  if (!(Atom3d_SqDistance(don,acc) <= dist*dist)){
    /*printf("Donor %d, Acceptor %d, rejected for distance\n",don->resnum,
      acc->resnum);*/
    return 0;
  }

  if (!(Atom3d_Angle(afp,acc,don) >= theta1)){
    /*printf("Donor %d, Acceptor %d, rejected for angle1\n",don->resnum,
      acc->resnum);*/
    return 0;
  }

  if (!(Atom3d_Angle(dfp,don,acc) >= theta2)){
    /*printf("Donor %d, Acceptor %d, rejected for angle2\n",don->resnum,
      acc->resnum);*/
    return 0;
  }

  if (!(fabs(Atom3d_Torsion(dfp,don,acc,afp)) >= torsion)){
    /*printf("Donor %d, Acceptor %d, rejected for torsion\n",don->resnum,
      acc->resnum);*/
    return 0;
  }

  return 1;
}

/***************************End backbone hbond utilities**********************/


/******************************************************************************
 *function: intra_peptide_hbonds                                              *
 *inputs:   atoms: pylist of atom3ds from a linusMol                          *
 *          donors: pylist of atoms indices for donors used: atoms[donors[i]] *
 *          acceptors: same as donors but for acceptors                       *
 *          donor_sat: pylist same length as donors containing a numeric code *
 *                     for hbond satisfaction.  -1 means unsatisfied.         *
 *                     -1000002 means N-terminus and                          *
 *                     -1000001 means proline nitrogen                        *
 *          acceptor_sat:  pylist same length as donors containing a          *
 *                         numeric code for hbond satisfaction.               *
 *                         -1 means unsatisfied.                              *
 *          donor_types: pylist specifying what type of hbond is made/donor   *
 *                       'NS' = Not Satisfied                                 *
 *                       'BB' = Backbone                                      *
 *                       'WAT'= Water                                         *
 *                       'SC' = Sidechain                                     *
 *          acceptor_types: same as donor_types but for acceptors             *
 *          radius: the maximum distance in SEQUENCE between 2 residues       *
 *          dist: the maximum distance between donor and acceptor             *
 *          theta1: the minimum angle between C-O-N                           *
 *          theta2: the minimum angle between C-N-O                           *
 *          torsion: the minimum angle between C-O-N-C                        *
 *                                                                            *
 *outputs: Information regarding bb hydrogen bonds are passed by reference    *
 *         into donor_sat and acceptor_sat (atom idx of hbond partner passed) *
 *     and into donor_types and acceptor_types as "BB".                       *
 *                                                                            *
 *method:    this function iterates though each backbone acceptor for each    *
 *unsatisfied backbone donor until it finds an hbond (determined by distance, *
 *           theta1, theta2, and torsion) or the loop completes.  If an hbond *
 *           is found, it records the partner indices in the donor_sat and    *
 *           acceptor_sat lists and "BB" into the donor_types and the         *
 *           acceptor_types lists.  It then searches through all unsatisfied  *
 *           backbone acceptors to see if any satisfied bb donors could make a*
 *           bifurcated hbond.  If any are found, results are stored in the   *
 *           acceptor_sat and acceptor_types lists only, since the donors were*
 *           already satisfied.                                               *
 *                                                                            *
 *Last modified: 8/28/07                                                      *
 *                                                                            *
 *****************************************************************************/

void intra_peptide_hbonds(PyObject *atoms, PyObject *donors, 
			  PyObject *acceptors,PyObject *donor_sat, 
			  PyObject *acceptor_sat, PyObject *donorTypes,
			  PyObject *acceptorTypes,int radius,double dist,
			  double theta1, double theta2,double torsion)
{

  Atom3d *don,*acc,*dfp,*afp;
  int i,j;
  int dsize = PyList_GET_SIZE(donors);
  int asize = PyList_GET_SIZE(acceptors);
  int satisfied;

  for (i=0; i<dsize; i++){
    don = GET_ATOM(atoms,GET_INT(donors,i));
    satisfied = GET_INT(donor_sat,don->resnum);
    
    /*If donor is already hbond satisfied, continue*/
    if (satisfied != -1)
      continue;

    if (don->resnum == 0 && strcmp(don->name," N  ")== 0 )  
      dfp = GET_ATOM(atoms,GET_INT(donors,i)+1);
    else       
      dfp = (Atom3d *)don->first_parent;

    for (j=0; j<asize; j++){
      acc = GET_ATOM(atoms,GET_INT(acceptors,j));
      afp = (Atom3d *)acc->first_parent;

      if (!hbonded(don,acc,dfp,afp,dist,theta1,theta2,torsion,radius))
	continue;

      /*If hbonded, record partners in donor and acceptor lists*/
      PyList_SET_ITEM(donor_sat,don->resnum,
		      PyInt_FromLong(GET_INT(acceptors,j)));
      PyList_SET_ITEM(acceptor_sat,acc->resnum,
		      PyInt_FromLong(GET_INT(donors,i)));
      PyList_SET_ITEM(donorTypes,don->resnum,PyString_FromString("BB"));
      PyList_SET_ITEM(acceptorTypes,acc->resnum,PyString_FromString("BB"));

      break;

    }
  }

  /*This checks for bifurcated N hbonds*/
  for (i=0; i<asize; i++){
    
    acc = GET_ATOM(atoms,GET_INT(acceptors,i));
    satisfied = GET_INT(acceptor_sat,acc->resnum);

    if (satisfied != -1)
      continue;

    afp = (Atom3d *)acc->first_parent;

    for (j=0; j<dsize; j++){

      don = GET_ATOM(atoms,GET_INT(donors,j));
      if (don->resnum == 0 && strcmp(don->name," N  ")== 0)  
	dfp = GET_ATOM(atoms,GET_INT(donors,i)+1);
      else       
	dfp = (Atom3d *)don->first_parent;

      /*If the donor is unsatisfied, it could not be an hbond partner        *
       *with the acceptor since it would have been found in the previous loop*
       *so... continue!                                                      */

      if (GET_INT(donor_sat,don->resnum) == -1)
	continue;

      if (!hbonded(don,acc,dfp,afp,dist,theta1,theta2,torsion,radius))
	continue;
      
      PyList_SET_ITEM(acceptor_sat,acc->resnum,
		      PyInt_FromLong(GET_INT(donors,j)));
      PyList_SET_ITEM(acceptorTypes,acc->resnum,PyString_FromString("BB"));
      break;
    }
  }

  return;
}

/**************************Begin water hbond utilities************************/

static int Atom3d_Close(Atom3d *a1, Atom3d *a2, double d)
{
	double dx, dy, dz;

	if (d < 0.0e0) {
		double r1= a1->radius;
		double r2= a2->radius;
		if (r1<1.0e-6 || r2<1.0e-6) return 0;
		d = r1 + r2;
	}
	dx = fabs(a1->x - a2->x);
	if (dx >= d) return 0;
	dy = fabs(a1->y - a2->y);
	if (dy >= d) return 0;
	dz = fabs(a1->z - a2->z);
	if (dz >= d) return 0;
	if ((dx*dx + dy*dy + dz*dz) < (d*d))
		return 1;
	else
		return 0;
}

void Atom3d_find_close(int close[],Atom3d *refatm, PyObject *atoms,int natoms,
		       int resdist, PyObject *afp)
{

  int i,startatm,endatm;
  int nres = PyList_GET_SIZE(afp)-1;
  Atom3d *curr_atm;


  startatm = GET_INT(afp,(int) max(0.0,(double)(refatm->resnum - resdist)));
  endatm   = GET_INT(afp,(int) min((double) nres, 
				   (double)(refatm->resnum+resdist)+1.0));

  for (i=startatm; i<endatm; i++){
    curr_atm = GET_ATOM(atoms,i);
    if ((strcmp(curr_atm->name," CA ") != 0))
      continue;
    if (Atom3d_Close(refatm,curr_atm,20.0)){
      close[curr_atm->resnum] = 1;
    }
  }
  return;
}

int Atom3d_has_bumps(int resdist, int close[], Atom3d *refatm, int hbatmnum,
		     PyObject *atoms, PyObject *first_atom_indices)
{
  int i,j,startatm,endatm;
  int nres     = PyList_GET_SIZE(first_atom_indices)-1;
  int startres = (int) max(0.0,(double)(refatm->resnum - resdist));
  int endres   = (int) min((double) nres,(double)(refatm->resnum+resdist)+1.0);
  Atom3d *curr_atm; 

  for (i=startres; i<endres; i++){

    if (close[i] == 0)
      continue;

    startatm = GET_INT(first_atom_indices,i);
    endatm   = GET_INT(first_atom_indices,i+1);

    for (j=startatm; j<endatm; j++){
      
      /*Don't include hbonding atom in bumpcheck*/
      if (j == hbatmnum)
	continue;
      
      curr_atm = GET_ATOM(atoms,j);
      /*printf("Atom %s %d  dist from water %d.  Dist: %f\n", curr_atm->name,
	       j, refatm->resnum,
	       Atom3d_Distance(refatm,curr_atm));*/
      if (Atom3d_Close(refatm,curr_atm,-1.0)){
	/*printf("Atom %s %d clashes with water %d.  Dist: %f\n", curr_atm->name,
	       j, refatm->resnum,
	       Atom3d_Distance(refatm,curr_atm));*/
	return (1);
      }
    }
  }
  return (0);
}



long int water_satisfied(int close[], int hbdatmnum, PyObject *atoms,
			 PyObject *wDists, PyObject *wAngles,
			 PyObject *wTorsions,int check_entropy,int resdist,
			 PyObject *afp, Atom3d *fp, Atom3d *sp, Atom3d *tp,
			 double water_radius)
{  
  int npos   = PyList_GET_SIZE(wDists);
  int natoms = PyList_GET_SIZE(atoms);
  int nres   = PyList_GET_SIZE(afp)-1;
  int i,j;
  int startres = (int) max(0.0,(double)(fp->resnum - resdist));
  int endres   = (int) min((double) nres,(double)(fp->resnum+resdist)+1.0);
  long pos = 0;
  Atom3d *wat = Atom3d_Create();

  /*Since wat is a dummy atom, deallocate its memory in python*/
  Atom3d_dealloc(wat);

  for (i=0; i<npos; i++){

    /*Assign wat its parents and cartesian coordinates from internal ones*/
    make_atom(fp,GET_FLT(wDists,i),sp,GET_FLT(wAngles,i),tp,
	      GET_FLT(wTorsions,i),wat);

    wat->resnum = fp->resnum;
    wat->radius = water_radius;

    Atom3d_find_close(close,wat,atoms,natoms,resdist,afp);
    for (j=startres; j<endres; j++){
      if (Atom3d_has_bumps(resdist, close, wat, hbdatmnum,atoms,afp))
	break;

      /*If no atoms within resdist bump with water, return/store results*/

      if (j==endres-1)
	if (!check_entropy)
	  return (long) pow(2,i);
       	else
	  pos += (long) pow(2,i);
    }
  }

  return pos;
}

/******************************End water hbond utilities**********************/


/******************************************************************************
 *function: water_hbonds                                                      *
 *inputs:   atoms: pylist of atom3ds from a linusMol                          *
 *          donors: pylist of atoms indices for donors used: atoms[donors[i]] *
 *          acceptors: same as donors but for acceptors                       *
 *          donor_sat: pylist same length as donors containing a numeric code *
 *                     for hbond satisfaction.  -1 means unsatisfied.         *
 *                     -1000002 means N-terminus and                          *
 *                     -1000001 means proline nitrogen                        *
 *          acceptor_sat:  pylist same length as donors containing a          *
 *                         numeric code for hbond satisfaction.               *
 *                         -1 means unsatisfied.                              *
 *          donorTypes : pylist specifying what type of hbond is made/donor   *
 *                       'NS' = Not Satisfied                                 *
 *                       'BB' = Backbone                                      *
 *                       'WAT'= Water                                         *
 *                       'SC' = Sidechain                                     *
 *          acceptorTypes : same as donor_types but for acceptors             *
 *          afp: Pylist of residue first atom indices                         *
 *          dDists: Pylist of water probe distances for donors                *
 *          dAngles:Pylist of water probe altitudinal angles for donors       *
 *          dTorsions: PyList of water probe azimuthal angles for donors      *
 *      *** Note: The dDists, dAngles and dTorsions MUST be the same size     *
 *          and all values are doubles***                                     *
 *          aDists: Pylist of water probe distances for acceptors             *
 *          aAngles:Pylist of water probe altitudinal angles for acceptors    *
 *          aTorsions: PyList of water probe azimuthal angles for acceptors   *
 *      *** Note: The aDists, aAngles and aTorsions MUST be the same size     *
 *          and all values are doubles***                                     *
 *          entropy_check: integer with value 1 (check all water positions)   *
 *                         or value 0 (find only 1 water that hbonds and quit)*
 *          resdist: abs(wat->resnum - atom->resnum) <= resdist in order for  *
 *                   a clash to be found                                      *
 *          water_radius: double for size of water probe.  Default: 1.4 A     *
 *          wat_mask: number used in donor_sat and acceptor_sat lists to      *
 *                    differentiate between waters and protein hbond partners *
 *                    this helps to clear up confusion about whether something*
 *                    is a sidechain or water.  Default value is -100000, and *
 *                    this value is added pos.  So if there are 5 waters that *
 *                    satisfy an hbond, and 5 possible waters, pos = 31.  The *
 *                    value saved in donor_sat or acceptor_sat is 31+-100000, *
 *                    so it would be -99969.                                  *
 *                                                                            *
 *outputs: Information regarding water hydrogen bonds is passed by reference  *
 *         into donor_sat and acceptor_sat (atom idx of hbond partner passed) *
 *     and into donor_types and acceptor_types as "WAT".                      *
 *                                                                            *
 *method:    this function tests unsatisfied bb donors and acceptors for water*
 *           hydrogen bonds by making a water probe and placing it in various *
 *           positions around the donor or acceptor.  These positions are     *
 *           specified by the Dists, Angles, and Torsions lists.  The water   *
 *           probe is placed in a position, and then the program checks if it *
 *           clashes with any atoms in any residues within resdist of the     *
 *           resnum of the donor/acceptor in question.  If entropy_check = 0, *
 *           the calculation over and the next unsatisfied donor/acceptor is  *
 *           checked.  If entropy_check = 1, all water probes are checked.    *
 *           All possible water positions correspond to a power of 2, so a    *
 *           decimal sum is returned through the donor_sat and acceptor_sat   *
 *           lists.  This number is converted to a binary list back in python,*
 *           which enables the number of hbonded waters and their distances,  *
 *           angles, and torsions to be determined.                           *
 *                                                                            *
 *Last modified: 8/28/07                                                      *
 *                                                                            *
 *****************************************************************************/
    
void water_hbonds(PyObject *atoms, PyObject *donors, PyObject *acceptors,
		  PyObject *donor_sat, PyObject *acceptor_sat,
		  PyObject *donorTypes,PyObject *acceptorTypes,
		  PyObject *afp, PyObject *dDists, PyObject *dAngles,
                  PyObject *dTorsions,PyObject *aDists, PyObject *aAngles,
                  PyObject *aTorsions, int entropy_check, int resdist,
		  double water_radius, int wat_mask) 
{

  Atom3d *fp,*sp,*tp,*curr_atm;
  int dsize =  PyList_GET_SIZE(donors);
  int asize =  PyList_GET_SIZE(acceptors);
  int natoms = PyList_GET_SIZE(atoms);
  int nres   = PyList_GET_SIZE(afp)-1;
  int i,j;
  int close[nres];
  long pos = 0;
  Atom3d *H = Atom3d_Create();

  set_list(close,nres,0);
  
  
  /*Check unsatisfied donors*/
  for (i=0; i<dsize; i++){
        
    set_list(close,nres,0);
    /**********Set parents for water*******/
    fp = GET_ATOM(atoms,GET_INT(donors,i));

    /*If current donor is already satisfied, move on*/
    if (GET_INT(donor_sat,fp->resnum) != -1)
      continue;
    
    /*Since the N-terminus doesn't have first and second parents, assign
      alternatives*/
    if (fp->resnum == 0 && strcmp(fp->name," N  ")== 0){
      sp = GET_ATOM(atoms,GET_INT(donors,i)+1);
      tp = GET_ATOM(atoms,GET_INT(donors,i)+2);
    }

    /*Otherwise use the parents that are available.*/
    else{
      sp = (Atom3d *) fp -> first_parent;
      tp = (Atom3d *) fp -> second_parent;
    }

    /*Build the hydrogen atom*/
    
    Atom3d_parents_INCREF(fp,sp,tp);
    
    make_atom(fp,1.008,sp,119.15,tp,0.0,H);
    H->resnum = fp->resnum;

    /*Check to see if donor can be water satisfied*/
    pos = water_satisfied(close,GET_INT(donors,i),atoms, dDists, dAngles, 
			  dTorsions,entropy_check, resdist, afp, H, fp, sp,
			  water_radius);

    /*If the donor can be water satisfied, mask its pos value by subtracting
     *1e5 and store it.  Pos values will be unmasked back in Python*/

    if (pos != 0){
      PyList_SET_ITEM(donor_sat,fp->resnum,PyInt_FromLong(pos+wat_mask));
      PyList_SET_ITEM(donorTypes,fp->resnum,PyString_FromString("WAT"));
    }

    /*Decrease reference counts of parents since the references will change 
     *when H changes*/

    Atom3d_parents_XDECREF(fp,sp,tp);
      
  }
  /*Delete hydrogen from Python memory*/
  PyMem_DEL(H);
  
  /*Now check unsatisfied acceptors*/
  for (i=0; i<asize; i++){

    set_list(close,nres,0);

    /*Get parents for water*/
    fp = GET_ATOM(atoms,GET_INT(acceptors,i));

     /*If current acceptor is already satisfied, move on*/
    if (GET_INT(acceptor_sat,fp->resnum) != -1)
      continue;

    sp = (Atom3d *) fp -> first_parent;
    tp = (Atom3d *) fp -> second_parent;

    pos = water_satisfied(close,GET_INT(acceptors,i),atoms, aDists, aAngles, 
			  aTorsions, entropy_check,resdist, afp, fp, sp, tp,
			  water_radius);
    if (pos != 0){
      PyList_SET_ITEM(acceptor_sat,fp->resnum,PyInt_FromLong(pos+wat_mask));
      PyList_SET_ITEM(acceptorTypes,fp->resnum,PyString_FromString("WAT"));
    }
  }
    
    
  return;
}

/*This function is not used in the LINUS code.  It finds an sc hbond partner by
  iterating through the list of donors and acceptors and stopping once the 
  first is found.  This cannot be reused, so it is marked accordingly.  
  Note: if a sidechain such as ASN is used, it can be an hbond partner twice:
  once with a donor, and once with an acceptor*/

void bb_sc_hbonds(PyObject *atoms,PyObject *bbList,PyObject *bbSat,
		  PyObject *scList,PyObject *scNames,PyObject *scParams)
{

  int bbLen = PyList_GET_SIZE(bbList);
  int scLen = PyList_GET_SIZE(scList);
  double dist;
  int used[scLen];
  Atom3d *BB, *SC;
  char *resname;
  int i, j;

  set_list(used,scLen,0);

  for (i=0; i<bbLen; i++){
    
    /*If the bb donor/acceptor is already hbond satisfied, move on*/
    if (GET_INT(bbSat,i) != -1)
      continue;

    BB = GET_ATOM(atoms,GET_INT(bbList,i));

    /*For each sidechain acceptor/donor*/
    for (j=0; j<scLen; j++){
      
      /*Get Sidechain atom, residue name, and max dist SC can be 
	from the BB atom*/
      SC = GET_ATOM(atoms,GET_INT(scList,j));

      resname = GET_STR(scNames,j);

      /*Get the maximum distance between the specific residue and its 
	hbond partner*/
      GET_DICT_DBL(scParams,resname,dist);

      
      /*If the distance between BB and SC <= max and the sidechain isn't
	satisfying an hbond elsewhere, count the bb donor/acceptor as satisfied
	and do not allow the sidechain to be used again*/

      if (Atom3d_Distance(BB,SC) <= dist && !used[j]){
	printf("Atom3d_Distance %d %d: %f\n",BB->resnum, SC->resnum,
	       Atom3d_Distance(BB,SC));
	used[j] = 1;
	PyList_SET_ITEM(bbSat,i,PyInt_FromLong(GET_INT(scList,j)));
	break;
      }

    }
  }
  return;
}


/******************************************************************************
 *function: bb_sc_sat                                                         *
 *inputs:   atoms: pylist of atom3ds from a linusMol                          *
 *          bbList: pylist of atoms indices for donors/acceptors.             *
 *          bbSat:     pylist same length as bbSat  containing a numeric code *
 *                     for hbond satisfaction.  -1 means unsatisfied.         *
 *                     -1000002 means N-terminus and                          *
 *                     -1000001 means proline nitrogen                        *
 *          scList: pylist of CA indices for possible sidechain dons/accs     *
 *          scNames:pylist of 3-letter residue codes for each element scList  *
 *       ***Note: scList and scNames must be the same size***                 *
 *           scParams: pyDict with maximum distance between sidechain CA and  *
 *                    hbond donor/acceptor                                    *
 *          scStart: integer for index to start searching scList              *
 *          scEnd:   integer for index to end searching scList                *
 *          scRad: maximum residue distance between bb donor/acceptor and sc  *
 *          sc_mask: number used in donor_sat and acceptor_sat lists to       *
 *                    differentiate between sidechain & protein hbond partners*
 *                    this helps to clear up confusion about whether something*
 *                    is a sidechain or water.  Default value is 1000000, and *
 *                    this value is added partners.  So if partners = 21, the *
 *                    value recorded in donor_sat or acceptor_sat = 1000021   *
 *                                                                            *
 *outputs: Information regarding sc    hydrogen bonds is passed by reference  *
 *         into donor_sat and acceptor_sat (partners+sc_mask is passed)       *
 *                                                                            *
 *method:    this function searches for bb-sc hbonds by measuring the distance*
 *           between the sidechain CA and the donor/acceptor.  CA was chosen  *
 *           because sidechain sampling is too difficult, so it is better to  *
 *           stick with backbone-backbone distances and assume the sidechain  *
 *           can satisfy the hbond.  All possible sidechains are tested for   *
 *           each unsatisfied donor/acceptor, and are stored in the variable  *
 *           "partners".  If partners != 0, partners+wat_mask is stored in the*
 *           donor_sat/acceptor_sat list.  Partners represents the positions  *
 *           of hbond satisfying sidechains, each of which corresponds to a   *
 *           binary number.  This number is converted to a binary list in     *
 *           Python, and the 1 sidechain is chosen per each donor/acceptor in *
 *           a way that allows the most donors/acceptors to be satisfied.     *
 *                                                                            *
 *Last modified: 8/28/07                                                      *
 *                                                                            *
 *****************************************************************************/

long bb_sc_sat(PyObject *atoms,PyObject *bbList,PyObject *bbSat, 
	       PyObject *scList,PyObject *scNames,PyObject *scParams,
	       int scStart, int scEnd, int scRad, int sc_mask)
{

  int bbLen = PyList_GET_SIZE(bbList);
  double dist;
  Atom3d *BB, *SC;
  char *resname;
  int i, j;
  long partners = 0;
  long satisfied = 1;
	
  for (i=0; i<bbLen; i++){
    
    partners = 0;

    BB = GET_ATOM(atoms,GET_INT(bbList,i));

    /*If the bb donor/acceptor is already hbond satisfied, move on*/
    if (GET_INT(bbSat,BB->resnum) != -1)
      continue;

    /*For each sidechain acceptor/donor*/
    for (j=scStart; j<scEnd; j++){

      
      /*Get Sidechain atom, residue name, and max dist SC can be 
	from the BB atom*/
      SC = GET_ATOM(atoms,GET_INT(scList,j));


      /*If SC is too far in sequence from BB, continue*/
      if (abs(SC->resnum - BB->resnum) > scRad)
	continue;

      resname = GET_STR(scNames,j);

      /*Get the maximum distance between the specific residue and its 
	hbond partner*/
      GET_DICT_DBL(scParams,resname,dist);

      
      /*If the distance between BB and SC <= max and the sidechain isn't
	satisfying an hbond elsewhere, count the bb donor/acceptor as satisfied
	and do not allow the sidechain to be used again*/

      if (Atom3d_Distance(BB,SC) <= dist){
	/*printf("Atom3d_Distance %d %d: %f\n",BB->resnum, SC->resnum,
	  Atom3d_Distance(BB,SC));*/
	partners += pow(2,j);
      }

    }

    /*If any donor/acceptor has no possible side chain partner, 
      return 0 when hbond check is complete*/
    if (partners == 0){ 
      satisfied = 0;
      continue;
    }

    PyList_SET_ITEM(bbSat,BB->resnum,PyInt_FromLong(partners+sc_mask));
  }
  return (satisfied);
}

    
/*************************Begin quick check utilities************************/ 

long water_quick_check(PyObject *atoms, 
		       PyObject *watDists, PyObject *watAngles, 
		       PyObject *watTors, PyObject *afp, 
		       Atom3d *fp, int check_entropy,
		       int atmnum, int resdist,
		       double water_radius)
{

  long sat = -1;
  Atom3d *H = Atom3d_Create();
  int nres = PyList_GET_SIZE(afp)-1;
  Atom3d *sp, *tp;
  
  int close[nres];

  set_list(close,nres,0);

  if (strcmp(" N  ",fp->name)==0){

      sp = (Atom3d *) fp -> first_parent;
      tp = (Atom3d *) fp -> second_parent;

      /*Build the hydrogen atom*/
    
      Atom3d_parents_INCREF(fp,sp,tp);
    
      make_atom(fp,1.008,sp,119.15,tp,0.0,H);
      H->resnum = fp->resnum;

      /*Check to see if donor can be water satisfied*/
      sat = water_satisfied(close, atmnum, atoms, watDists, watAngles, 
			    watTors,check_entropy, resdist, afp, H, fp, sp,
			    water_radius);
  }
  else{

       sp = (Atom3d *) fp -> first_parent;
       tp = (Atom3d *) fp -> second_parent;

       sat = water_satisfied(close,atmnum, atoms, watDists, watAngles,
			     watTors, check_entropy,resdist, afp, fp, sp,
			     tp,water_radius);
  }

  /*Remove H from Python memory*/

  Atom3d_dealloc(H);

  return sat;
}

long bb_quick_check(PyObject *atoms, double dist,double angle1,double angle2,
		    double torsion, int atmidx1,int atmidx2, int radius)
{
  long satlist_value = -1;
  Atom3d *don, *dfp, *acc, *afp,*atm;

  atm = GET_ATOM(atoms,atmidx1);

  if (strcmp(atm->name," N  ")== 0){
    
    don = GET_ATOM(atoms,atmidx1);
    dfp = don->first_parent;
    acc = GET_ATOM(atoms,atmidx2);
    afp = acc->first_parent;

  }
  else if (strcmp(atm->name," O  ")== 0){

    don = GET_ATOM(atoms,atmidx2);
    dfp = don->first_parent;
    acc = GET_ATOM(atoms,atmidx1);
    afp = acc->first_parent;
  }
  else{
    printf("Error in bb_quick_check.  Is not making a bb hbond\n");
  }

  if (hbonded(don,acc,dfp,afp,dist,angle1,angle2,torsion,radius))
    satlist_value = atmidx2;

  return satlist_value;
}
    
long sc_quick_check(PyObject *atoms, PyObject *residue_names, 
		    PyObject *distDict, int bbAtmNum, int scAtmNum)
{
  Atom3d *bb, *sc;
  char *resname;
  double dist;
  long satisfied = -1;

  bb = GET_ATOM(atoms,bbAtmNum);
  sc = GET_ATOM(atoms,scAtmNum);

  resname = GET_STR(residue_names,sc->resnum);
  GET_DICT_DBL(distDict,resname,dist);

  if (Atom3d_Distance(bb,sc)<dist)
    satisfied = scAtmNum;

  return satisfied;
}

int partner_in_interval(PyObject *atoms, int atomidx, int minres, int maxres)
{

  Atom3d *partner, *minatom, *maxatom;

  partner = GET_ATOM(atoms,atomidx);
  minatom = GET_ATOM(atoms,minres);
  maxatom = GET_ATOM(atoms,maxres);

  return (partner->resnum >= minatom->resnum && 
	  partner->resnum <= maxatom->resnum);

}

  

/****************************End quick check utilities************************/


/******************************************************************************
 *function: bb_sc_sat                                                         *
 *inputs:   atoms: pylist of atom3ds from a linusMol                          *
 *          hbondList: pylist of atoms indices for donors/acceptors.          *
 *          satList: pylist same length as hbondSat containing a numeric code *
 *                     for hbond satisfaction.  -1 means unsatisfied.         *
 *                     -1000002 means N-terminus and                          *
 *                     -1000001 means proline nitrogen                        *
 *          tmpSatList: satList from the last LINUS move                      *
 *          typeList: pylist of Hbond types.                                  *
 *                       'NS' = Not Satisfied                                 *
 *                       'BB' = Backbone                                      *
 *                       'WAT'= Water                                         *
 *                       'SC' = Sidechain                                     *
 *          tmpTypeList: typeList from last LINUS move                        *
 *          residue_names:pylist of 3-letter residue codes for each           *
 *                        sidechain donor/acceptor                            *
 *          distDict: pyDict with maximum distance between sidechain CA and   *
 *                    hbond donor/acceptor                                    *
 *          watDists: Pylist of water probe distances for donors/acceptors    *
 *          watAngs:Pylist of water probe altitudinal angles for donors/accep *
 *          watTors: PyList of water probe azimuthal angles for donors/accep  *
 *          first_atom_indices:  Pylist of residue first atom indices         *
 *          bbDist: the maximum distance between bb donor and acceptor        *
 *          bbTheta1: the minimum angle between bb-bb C-O-N                   *
 *          bbTheta2: the minimum angle between bb-bb C-N-O                   *
 *          bbTorsion: the minimum angle between bb-bb C-O-N-C                *
 *          water_radius: double for size of water probe.  Default: 1.4 A     *
 *          startidx:integer for start index to do quick check                *
 *          resdist: abs(wat->resnum - atom->resnum) <= resdist in order for  *
 *                   a clash to be found                                      *
 *          check_entropy: integer with value 1 (check all water positions)   *
 *                         or value 0 (find only 1 water that hbonds and quit)*
 *          wat_mask: number used in donor_sat and acceptor_sat lists to      *
 *                    differentiate between waters and protein hbond partners *
 *                    this helps to clear up confusion about whether something*
 *                    is a sidechain or water.  Default value is -100000, and *
 *                    this value is added pos.  So if there are 5 waters that *
 *                    satisfy an hbond, and 5 possible waters, pos = 31.  The *
 *                    value saved in donor_sat or acceptor_sat is 31+-100000, *
 *                    so it would be -99969.                                  *
 *                                                                            *
 *outputs: Information regarding hydrogen bonds is passed by reference        *
 *         into satList (with waters masked by wat_mask)                      *
 *         and into typeList (same codes as given above)                      *
 *                                                                            *
 *method:    this function uses hbond satisfaction information from the       *
 *           last LINUS move to see if any hbonds were satisfied the same way.*
 *           This saves time when doing the full-fledged hbond calculations to*
 *           follow.  All methods are the same as in previous functions, but  *
 *           only the specific hbond type is checked.  That is, if N 31 in the*
 *           last LINUS move hbonded to O 35, only that hbond would be checked*
 *           Same with waters and sidechain partners.  If the hbonds are still*
 *           satisfied, the satList and tyleList is updated accordingly.      *
 *                                                                            *
 *Last modified: 8/28/07                                                      *
 *                                                                            *
 *****************************************************************************/
  
void quick_check(PyObject *atoms, PyObject *hbondList,PyObject *satList,
		 PyObject *tmpSatList, PyObject *typeList,
		 PyObject *tmpTypeList,
		 PyObject *residue_names, PyObject *distDict,
		 PyObject *watDists,PyObject *watAngs, PyObject *watTors,
		 PyObject *first_atom_indices,
		 double bbDist, double bbTheta1, double bbTheta2,
		 double bbTorsion,double water_radius,
		 int startidx, int endidx,int resdist,int check_entropy,
		 int wat_mask)
{

  int i, minres,maxres;
  long sat = 0;
  char *type;
  long new_result;
  Atom3d *tmp1, *tmp2;
  
  for (i=startidx;i<endidx;i++){

    sat = GET_INT(tmpSatList,i);

    if (sat == -1) continue;

    type = GET_STR(tmpTypeList,i);
    switch(type[0]){
      
      case 'W':
	new_result = water_quick_check(atoms,
				       watDists,watAngs,watTors,
				       first_atom_indices,
				       GET_ATOM(atoms,GET_INT(hbondList,i)),
				       check_entropy,GET_INT(hbondList,i),
				       resdist,water_radius);
	if (new_result != 0){
	  PyList_SET_ITEM(satList,i,PyInt_FromLong(new_result+wat_mask));
	  PyList_SET_ITEM(typeList,i,PyString_FromString("WAT"));
	}
	break;
      case 'B':
	if (!partner_in_interval(atoms,GET_INT(tmpSatList,i),
				 GET_INT(hbondList,startidx),
				 GET_INT(hbondList,endidx-1)))
	  break;

	new_result = bb_quick_check(atoms,bbDist,bbTheta1,bbTheta2,bbTorsion,
				    GET_INT(hbondList,i),GET_INT(tmpSatList,i),
				    resdist);
	if (new_result != -1){
	  PyList_SET_ITEM(satList,i,PyInt_FromLong(new_result));
	  PyList_SET_ITEM(typeList,i,PyString_FromString("BB"));
	}
	break;
      case 'S':
	if (!partner_in_interval(atoms,GET_INT(tmpSatList,i),
				 GET_INT(hbondList,startidx),
				 GET_INT(hbondList,endidx-1)))
	  break;

	new_result = sc_quick_check(atoms, residue_names, distDict,
				   GET_INT(hbondList,i),GET_INT(tmpSatList,i));
	if (new_result != -1){
	  PyList_SET_ITEM(satList,i,PyInt_FromLong(new_result));
	  PyList_SET_ITEM(typeList,i,PyString_FromString("SC"));
	} 
	break;
      default:
	continue;
    }
  }
}


/**************************Begin Python Wrapper*******************************/

PyObject* py_make_atom(PyObject *unused, PyObject *args)
{
  Atom3d *fp = (Atom3d *)       PyTuple_GET_ITEM(args,0) ;
  double  d  = PyFloat_AsDouble(PyTuple_GET_ITEM(args,1));
  Atom3d *sp = (Atom3d *)       PyTuple_GET_ITEM(args,2) ;
  double  a  = PyFloat_AsDouble(PyTuple_GET_ITEM(args,3));
  Atom3d *tp = (Atom3d *)       PyTuple_GET_ITEM(args,4) ;
  double  t  = PyFloat_AsDouble(PyTuple_GET_ITEM(args,5));
  Atom3d *at = (Atom3d *)       PyTuple_GET_ITEM(args,6) ;

  Atom3d_parents_XDECREF(at->first_parent,at->second_parent,at->third_parent);

  Atom3d_parents_INCREF(fp,sp,tp);
  
  make_atom(fp,d,sp,a,tp,t,at);

  Py_INCREF(Py_None);
  return Py_None;
}


/*This function doesn't work.  I wrote it early in the project and didn't
 *get a chance to fix it.  Maybe later...*/
PyObject * py_add_waters(PyObject *unused, PyObject *args)
{

  PyObject *dList, *aList, *dWaters, *aWaters, *dDist, *aDist;
  PyObject *dAng, *aAng, *dTors, *aTors, *atoms;
  PyObject *aPos = NULL;
  PyObject *dPos = NULL;
  Atom3d *water;
  int i;
  int j = 0;

  if (!PyArg_ParseTuple(args,"OOOOOOOOOOOOO|OO", &dList, &aList, &dWaters, 
			&aWaters, &dDist, &aDist, &dAng, &aAng,
			&dTors, &aTors, &atoms,&dPos, &aPos)) 
    return NULL;

  /*place_donor_wat   (dlist,dWaters,dDist,dAng,dTors,atoms,dPos);
    place_acceptor_wat(alist,aWaters,aDist,aAng,aTors,atoms,aPos);*/
  
  Py_INCREF(Py_None);
  return Py_None;
}

PyObject * py_intra_peptide_hbonds(PyObject *unused, PyObject *args)
{

  PyObject *atoms, *donors, *acceptors, *donor_sat, *acceptor_sat;
  PyObject *donorTypes, *acceptorTypes;
  int radius;
  double dist, theta1, theta2, torsion;
  
  if (!PyArg_ParseTuple(args,"OOOOOOOidddd", &atoms, &donors, &acceptors,
			&donor_sat, &acceptor_sat, &donorTypes, &acceptorTypes,
			&radius, &dist, &theta1, &theta2, &torsion)) 
    return NULL;

  intra_peptide_hbonds(atoms,donors,acceptors,donor_sat,acceptor_sat,
		       donorTypes,acceptorTypes,radius,
		       dist, theta1, theta2, torsion);

  Py_INCREF(Py_None);
  return Py_None;
}

PyObject * py_water_hbonds(PyObject *unused, PyObject *args)
{
  PyObject *atoms, *donors, *acceptors, *donor_sat, *acceptor_sat,*fai;
  PyObject *dDists, *dAngles, *dTorsions,*aDists, *aAngles, *aTorsions;
  PyObject *donorTypes, *acceptorTypes;
  int nres,entropy_check,resdist,wat_mask;
  double water_radius;

  if (!PyArg_ParseTuple(args, "OOOOOOOOOOOOOOiidi", &atoms, &donors,
			&acceptors,
			&donor_sat, &acceptor_sat, &donorTypes, &acceptorTypes,
			&fai, &dDists,
			&dAngles,&dTorsions,&aDists,&aAngles,&aTorsions,
			&entropy_check,&resdist,&water_radius,&wat_mask))
    return NULL;

  water_hbonds(atoms,donors,acceptors,donor_sat,acceptor_sat,
	       donorTypes,acceptorTypes,fai,
	       dDists,dAngles,dTorsions,aDists,aAngles,aTorsions, 
	       entropy_check,resdist,water_radius,wat_mask);

  Py_INCREF(Py_None);
  return Py_None;
}

PyObject * py_bb_sc_hbonds(PyObject *unused, PyObject *args)
{
  PyObject *atoms,*donors,*acceptors,*donor_sat,*acceptor_sat;
  PyObject *scDonors, *scAcceptors, *scDonorNames, *scAcceptorNames;
  PyObject *scParams;

  if (!PyArg_ParseTuple(args, "OOOOOOOOOO", &atoms, &donors, &acceptors,
			&donor_sat, &acceptor_sat, &scDonors, &scAcceptors,
			&scDonorNames,&scAcceptorNames,&scParams))

    return NULL;

  bb_sc_hbonds(atoms,donors,donor_sat,scAcceptors,scAcceptorNames,
	       scParams);
  bb_sc_hbonds(atoms,acceptors,acceptor_sat,scDonors,scDonorNames,
	       scParams);

  Py_INCREF(Py_None);
  return Py_None;
}  

PyObject * py_bb_sc_sat(PyObject *unused, PyObject *args)
{
  PyObject *atoms,*donors,*acceptors,*donor_sat,*acceptor_sat;
  PyObject *scPartners, *scPartnerNames, *scParams;
  int accEnd,donStart,scRad,sc_mask;
  int list_len;
  long donHbonded, accHbonded;

  if (!PyArg_ParseTuple(args, "OOOOOOOOiiii", &atoms, &donors, &acceptors,
			&donor_sat, &acceptor_sat,
			&scPartners,&scPartnerNames,&scParams,
			&accEnd,&donStart,&scRad,&sc_mask))

    return NULL;

  list_len = PyList_GET_SIZE(scPartners);

  donHbonded= bb_sc_sat(atoms,donors,donor_sat,
			scPartners,scPartnerNames,scParams,0,accEnd,scRad,
			sc_mask);


  accHbonded= bb_sc_sat(atoms,acceptors,acceptor_sat,
			scPartners,scPartnerNames,scParams,donStart,list_len,
			scRad,sc_mask);

  if ((!accHbonded) || (!donHbonded)) return PyInt_FromLong(0L);

  return PyInt_FromLong(1L);
}  

PyObject * py_quick_check(PyObject *unused, PyObject *args)
{

  PyObject *atoms, *hbondList,*satList,*tmpSatList,*typeList, *tmpTypeList;
  PyObject *residue_names, *distDict,*watDists,*watAngs,*watTors;
  PyObject *first_atom_indices;
  int startidx, endidx, check_entropy,resdist,wat_mask;
  double bbDist, bbTheta1, bbTheta2, bbTorsion,water_radius;

  if (!PyArg_ParseTuple(args, "OOOOOOOOOOOOiiiidddddi", &atoms, &hbondList,
			&satList,&tmpSatList, &typeList, &tmpTypeList,
			&residue_names,
			&distDict, &watDists, &watAngs, &watTors,
			&first_atom_indices,
			&startidx,&endidx,&check_entropy,&resdist,
			&bbDist,&bbTheta1,
			&bbTheta2,&bbTorsion,&water_radius,&wat_mask))
    return NULL;

  quick_check(atoms,hbondList,satList,tmpSatList,typeList,tmpTypeList,
	      residue_names,
	      distDict,watDists,watAngs,watTors,first_atom_indices,
	      bbDist,bbTheta1, bbTheta2, 
	      bbTorsion,water_radius,startidx,endidx,resdist,check_entropy,
	      wat_mask);

  Py_INCREF(Py_None);
  return Py_None;
}
  
static PyMethodDef moduleMethods[] = 
  {
    {"make_atom", py_make_atom, METH_VARARGS,
     "make_atom(fp,d,sp,a,tp,t,at) -> atom3d\n\n"
     "Makes a new atom from internal coordinates and specified neighbors"
    },

    {"add_waters",py_add_waters,METH_VARARGS,
     "add_waters(linusMol,water_list,numatoms) -> void\n\n"
     "Gives coordinates to a list of waters with no prior information given."
    },

    {"intra_peptide_hbonds",py_intra_peptide_hbonds,METH_VARARGS,
     "intra_peptide_hbonds(atomlist,donors,acceptors,donor_sat,acceptor_sat,radius,dist,theta1,theta2,torsion) -> void\n\n"
     "Finds intrapeptide hbonds within a specific radius with user-specified hbond distance, angles, and torsion."
    },

    {"water_hbonds",py_water_hbonds,METH_VARARGS,
     "water_hbonds(atomlist,donors,acceptors,donor_sat,acceptor_sat,pname,nres,residue_first_atom_indices)->void\n\n"
     "Checks waters as possible hbond donors or acceptors."
    },

    {"bb_sc_hbonds",py_bb_sc_hbonds,METH_VARARGS,
     "bb_sc_hbonds(atoms,donors,acceptors,donor_sat,acceptor_sat,scDonors,scAcceptors,scDonorNames,scAcceptorNames,scDonorParams,scAcceptorParams->void\n\n"
     "Checks for backbone-sidechain satisfaction by measuring the distance between the backbone donor/acceptor and sidechain CA atom."
    },
    
    {"bb_sc_sat",py_bb_sc_sat,METH_VARARGS,
     "bb_sc_sat(atoms, donors, acceptors, donor_sat, acceptor_sat, scPartners,scPartnerNames, scParams, accEnd, donStart)->void\n\n"
     "Checks for backbone-sidechain satisfaction allowing only a sidechain to satisfy only one hydrogen bond.  Returns the information in the sat lists.  Python contains a script for assigning the best side chain to each donor/acceptor (so that the max number of bb partners can be satisfied."
    },

    {"quick_check",py_quick_check,METH_VARARGS,
     "quick_check(atoms,satList,tmpSatList,typeList,residue_names,distDict,watDists,watAngs,watTors,bbDist,bbTheta1, bbTheta2, bbTorsion,startidx,endidx,is_nitrogen)->void\n\n"
     "Sees if donors/acceptors are satisfied in the same way as before the last move"
    },
    
    {NULL, NULL}
    
  };

/***********************************End Python Wrapper************************/


/*Atom3d.c functions cannot be used without this*/
PyMODINIT_FUNC inithbond(void)
{
  PyObject *m;
#ifndef _ATOM3D_STATIC
	import_atom3d();
#endif
    m = Py_InitModule4("hbond",                       /* Name of the module */
		   moduleMethods,                  /* Method table */
		   "A module containing various "  /* Module docstring */
		   "functions for computing hbonds.",
		   NULL,                           /* Doesn't change */
		   PYTHON_API_VERSION);            /* Doesn't change */
  return;
}
