"""This module defines two classes *LinusFiles* and *Linus*"""

import linusMol
import linusRun
from linusScore import *
from add_waters import *
import os
#from find_hbonds import aDists, aAngs, aTors, dDists, dAngs, dTors,hbondParams

default_simpar = {
    'beta': 2.0,
    'numcycle': 1000,
    'numsave': 10,
    'trials': 1,
    'updateweights': 0,
    'reset_conformation': 0,
    'use_ms_sampling': 0,
    'write_rc': 0
}

default_hbdpar = {
    'use_hbond': 0,
    'use_sidechain_hbond': 0,
    'hbond_distance': 3.5,
    'hbond_probe': 1.5,
    'hbond_score_short': 0.5,
    'hbond_score_long': 1.0,
    'hbond_torsion': 140.0,
    'sidechain_hbond_distance': 3.5,
    'sidechain_hbond_score': 1.0,
    'sidechain_hbond_torsion': 140.0,
    'hbond_winmin': 2,
    'hbond_winmax': 6
}

default_conpar = {
    'use_contact': 0,
    'contact_probe': 2.8,
    'contact_winmin': 2,
    'contact_winmax': 6,
    'contact_scale': 1.0
}

default_torpar = {
    'use_torsion': 0,
    'torsion_penalty': 1.0,
}

default_compar = {
    'use_compact': 0,
    'compact_fraction': 1.0
}

default_slvpar = {
    'use_solvation': 0,
    'solvation_probe': 1.4,
    'cavity_term': 0.0,
    'solvation_winmax': 6
}

converters = {
    type(0.0): float,
    type(0): int,
    type(''): str
}

StringType = type('')
IntType = type(0)

class LinusFiles:
    """Class to create files that are output by LINUS.  All filenames
    use the base name of the input file and then add the appropriate
    extensions to create new files.

    Methods of this class

        o *get_logfile* - returns a reference to the LOG file

        o *get_out_file* - returns a reference to the file used for saving
        structures during a simulation

        o *get_min_file* - returns a reference to the file used for saving
        the best energy structure found during a simulation

        o *get_rc_file* - returns a reference to the file used for saving
        the mesostate conformation strings found during a simulation

        o *get_swt_file* - returns *name* of the file used for saving
        the observed distribution of the residues in the various
        secondary structure states

        o *get_ps_file* - returns *name* of the file to which the
        postscript plot of the the observed distribution of the residues in
        the various secondary structure states is written

        o *create_file_name* - given an extension, creates the *name* of
        the file
    """
    
    def __init__(self, pdbfile):
        self.input = os.path.expanduser(pdbfile)
        self.basename = os.path.splitext(self.input)[0]
        self.log_filename = self.basename + '.LOG'
        self.out_filename = self.basename + '.out'
        self.min_filename = self.basename + '.min'
        self.rc_filename = self.basename + '.rc'

    def get_logfile(self):
        if hasattr(self, 'log_file'):
            return self.log_file
        else:
            self.log_file = open(self.log_filename, 'wb')
            return self.log_file

    def get_out_file(self):
        if hasattr(self, 'out_file'):
            return self.out_file
        else:
            self.out_file = open(self.out_filename, 'wb')
            return self.out_file

    def get_min_file(self):
        return self.min_filename
        if hasattr(self, 'min_file'):
            self.min_file.close()
        self.min_file = open(self.min_filename, 'wb')
        return self.min_file

    def get_rc_file(self):
        if not hasattr(self, 'rc_file'):
            self.rc_file = open(self.rc_filename, 'wb')
        return self.rc_file

    def get_swt_filename(self, wsiz, trial=1):
        from string import zfill
        wsiz = zfill(str(wsiz), 3)
        trial = zfill(str(trial), 3)
        swt_file = self.basename +'_' + wsiz +'_' + trial +'.SWT'
        return swt_file

    def get_ps_filename(self, wsiz, trial=1):
        from string import zfill
        wsiz = zfill(str(wsiz), 3)
        trial = zfill(str(trial), 3)
        ps_file = self.basename +'_' + wsiz +'_' + trial +'.ps'
        return ps_file

    def create_filename(self, ext, trial=1):
        from string import zfill
        trial = zfill(str(trial), 3)
        return self.basename + trial + '.' + ext
    
    

class Linus:
    """Class to describe a LINUS simulation object

    Instantiation

        An instance is created by calling the class constructor with
        two arguments - name of the pdb file and a random number generator
        The pdb file must reside in a directory to which the USER has
        write priviledges.

    Attributes

        o *protein* - instance of LinusMol - the protein that is being
        simulated

        o *simpar* - dictionary - simulation parameters. The various
        parameters are

            i.    *beta* - float-  reciprocal of the simulation temperature
            
            ii.   *numcycle* - integer - number of cycles of simulation to run
            
            iii.  *numsave*  - integer - frequency with which structures are
                  saved to disk
                              
            iv.   *reset_conformation* - boolean - if true at the end of
                  the simulation the molecules' conformation is reset
                  to the one defined in the input pdb file
                  
            v.    *update_weights* - boolean - if true at the end of the
                  simulation the sampling weights of the residues are
                  updated to the observed distribution in the various
                  secondary structural states during the course of the
                  simulation

        o *hbdpar* - hydrogen bonding energy parameters.

            i.   *use_hbond* - boolean - if true backbone - backbone
                 hydrogen bonds are enabled

            ii.  *use_sidechain_hbond* - boolean - if true sidechain
                 acceptor to backbone donor hydrogen bonds are
                 enabled

            iii. *hbond_distance* - float - optimal distance between
                 backbone donor and backbone acceptor

            iv.  *hbdond_torsion* - float - minium torsion angle between
                 the backbone acceptor and the backbone donor and it's two
                 antecedent atoms( 'C' of previous residue and 'CA' of the
                 residue

            v.   *hbond_probe* - float - distance over which the hydrogen
                 bond energy for backbone donor to backbone acceptor
                 bonds will be scaled to zero from the maximum value

            vi.  *hbond_score_short* - float - backbone donor to backbone
                 acceptor hydrogen bond energy for residues separations of
                 5 or less

            vii. *hbond_score_long* - float - backbone donor to backbone
                 acceptor hydrogen bond energy for residues separations of
                 6 or less

            viii. *sidechain_hbond_distance* - float - optimal distance between
                  backbone donor and sidechain acceptor

            ix.  *sidechain_hbond_torsion* - float - minium torsion angle
                 between the sidechain acceptor and the backbone donor and
                 it's two antecedent atoms ( 'C' of previous residue and 'CA' 
                 of the residue)

            x.   *sidechain_hbond_score* - float - energy of a sidechain
                 to backbone hydrogen bond

            xi.  *hbond_winmin* - integer - minimum separation between
                 two residues that can participate in backbone to backbone
                 hydrogen bonds

            xii. *hbond_winmax* - integer - maximum  separation between
                 two residues that can participate in backbone/sidechain
                 acceptor to backbone donor hydrogen bonds

        o *conpar* - contact energy parameters

            i.   *use_contact* - boolean - if true contact energy
                 calculations are enabled

            ii.  *contact_winmin* - integer - minimum sequence separation
                 between pairs of residues that participate in contact
                 energy calculations

            iii. *contact_winmax* - integer - maximum sequence separation
                 between pairs of residues that participate in contact
                 energy calculations

            iv.  *contact_probe* - float - distance over which the
                 contact energy is scaled to zero from the maximum value

            v.   *contact_scale* - float - scaling factor to apply to
                 the contact energy

        o *torpar* - torsion penalty score parameters

            i.   *use_torsion* - boolean - if true torsion energy calculation
                 is enabled

            ii.  *torsion_penalty* - float - penalty for adoption positive
                 phi values - for GLY this is a reward instead of a penalty

        o *files* - instance of LinusFiles - class to create various files
          for logging etc. during a simulation

        o *generator* - object - a random number generator

        o *registered functions* - list - list of functions to be run
          before, during and after a simulation
          

    Methods

        o *set_simulation_parameters* - set some or all of the
        simulation parameters

        o *set_hbond_parameters* - set some of all of the hydrogen
        bond parameters

        o *set_contact_parameters* - set some of all of the contact
        energy parameters

        o *set_torsion_parameters* - set parameters for torsion
        energy calculations

        o *set_weights* - set sampling weights for specified move type
        and specified list of residues

        o *normalize_weights* - normalize sampling weights so that
        the sum of the weights for each residue adds to 1.0

        o *read_weights_from_file* - read sampling weights for each
        residue from a file

        o *update_weights* - assign sampling weights for all residues
        from a python list of weights

        o *total_score* - total energy of a conformation

        o *load_mesostates* - load allowed mesostates for each residue
        from a file

        o *register_functions* - add user defined functions to be run on
        the molecule before, during or after the simulation

        o *run* - run the simulation
            
    """

    def __init__(self, pdbfile, generator):
        """Create a LINUS simulation object

            pdbfile - file containing coordinates of the
                      molecule to be simulated in PDB format

            generator - a random number generator.

        """
        
        pdbfile = os.path.expanduser(pdbfile)
        self.protein = linusMol.LinusProtein(pdbfile)
        self.simpar = default_simpar.copy()
        self.hbdpar = default_hbdpar.copy()
        self.conpar = default_conpar.copy()
        self.torpar = default_torpar.copy()
        self.compar = default_compar.copy()
        self.simpar['movelist'] = [0] + [1]*(self.protein.num_residues-2) + [0]
        self.slvpar = default_slvpar.copy()
        #self.hbondParams = hbondParams.copy()
        self.hblist = self.conlist = None
        self.generator = generator
        self.files = LinusFiles(pdbfile)
        self.trial = 0
        self.registered_functions = []

    def __getattr__(self, name):
        if self.__dict__.has_key(name):
            return self.__dict__[name]
        elif self.simpar.has_key(name):
            return self.simpar[name]
        elif self.hbdpar.has_key(name):
            return self.hbdpar[name]
        elif self.conpar.has_key(name):
            return self.conpar[name]
        elif self.torpar.has_key(name):
            return self.torpar[name]
        elif self.compar.has_key(name):
            return self.compar[name]
        elif self.slvpar.has_key(name):
            return self.slvpar[name]
        elif name == 'seeds':
            return self.generator.get_seed()
        else:
            raise AttributeError, 'Undefined attribute %s' % name

    def __setattr__(self, name, value):
        if name in ('pdbfile', 'protein', 'simpar', 'hbdpar', 'torpar',
                    'conpar', 'compar', 'hblist', 'conlist', 'trial', 'slvpar'):
            self.__dict__[name] = value
        elif not hasattr(self, name):
            self.__dict__[name] = value
        else:
            raise AttributeError, 'Attribute %s cannot be set directly' % name

    def set_simulation_parameters(self, **kw):
        """Set the simulation parameters

        Arguments

            series of keyword=value arguments

        Errors

            will raise AttributeError if invalid parameter name is specified
            and TypeError if invalid value for the parameter is supplied

        """    
        for key, value in kw.items():
            if self.simpar.has_key(key):
                vtype = type(self.simpar[key])
                self.simpar[key] = converters[vtype](value)
            else:
                raise KeyError, 'Undefined simulation parameter %s' % key


    def set_hbond_parameters(self, **kw):
        """Set the hydrogen bond parameters

        Arguments

            series of keyword=value arguments

        Errors

            will raise AttributeError if invalid parameter name is specified
            and TypeError if invalid value for the parameter is supplied

        """    
        
        for key, value in kw.items():
            if self.hbdpar.has_key(key):
                vtype = type(self.hbdpar[key])
                self.hbdpar[key] = converters[vtype](value)
            else:
                raise KeyError, 'Undefined hbond parameter %s' % key

    def set_contact_parameters(self, **kw):
        """Set the contact energy parameters

        Arguments

            series of keyword=value arguments

        Errors

            will raise AttributeError if invalid parameter name is specified
            and TypeError if invalid value for the parameter is supplied

        """    
        
        for key, value in kw.items():
            if self.conpar.has_key(key):
                vtype = type(self.conpar[key])
                self.conpar[key] = converters[vtype](value)
            else:
                raise KeyError, 'Undefined contact parameter %s' % key

    def set_torsion_parameters(self, **kw):
        """Set the torsion energy parameters

        Arguments

            series of keyword=value arguments

        Errors

            will raise AttributeError if invalid parameter name is specified
            and TypeError if invalid value for the parameter is supplied

        """    
        
        for key, value in kw.items():
            if self.torpar.has_key(key):
                vtype = type(self.torpar[key])
                self.torpar[key] = converters[vtype](value)
            else:
                raise KeyError, 'Undefined torsion parameter %s' % key

    def set_compaction_parameters(self, **kw):
        """Experimental - do not use this method"""
        
        for key, value in kw.items():
            if self.compar.has_key(key):
                vtype = type(self.compar[key])
                self.compar[key] = converters[vtype](value)
            else:
                raise KeyError, 'Undefined compaction parameter %s' % key

    def set_solvation_parameters(self, **kw):
        """Experimental - do not use this method"""
        
        for key, value in kw.items():
            if self.slvpar.has_key(key):
                vtype = type(self.slvpar[key])
                self.slvpar[key] = converters[vtype](value)
            else:
                raise KeyError, 'Undefined solvation parameter %s' % key
            
    def set_hbs_parameters(self, **kw):
        """Experimental - do not use this method"""
        
        for key, value in kw.items():
            if self.hbondParams.has_key(key):
                if isinstance(self.hbondParams[key],list):
                    self.hbondParams[key] = (value)
                else:
                    vtype = type(self.hbondParams[key])
                    self.hbondParams[key] = converters[vtype](value)
            else:
                raise KeyError, 'Undefined solvation parameter %s' % key

    def set_weights(self, move_type, weight, *res):
        """Set sampling weights for a specific move type

        Arguments

            o *move_type* - string - one of helix, strand, turn1, turn2
              and coil

            o *weight* - float - weight to use

            o **res* - This can take 3 forms:

                i. indices of the residues

                ii. names of residues

                iii. 'all' to set for all residues

        Usage

            sim.set_weights('helix', 0.0, 'all')

            will set the helix weight to 0 for all residues

            sim.set_weights('strand', 0.0, 1, 5, 9, 11)

            will set the strand weights to 0 for residues 1, 5, 9 and 11

            sim.set_weights('coil', 2.0, 'gly', 'pro')

            will set the coil weights to 2.0 for all glycines and prolines
        """
        
        if move_type == 'helix':
            w = self.protein.residue_helix_weights
        elif move_type== 'strand':
            w = self.protein.residue_strand_weights
        elif move_type == 'coil':
            w = self.protein.residue_coil_weights
        elif move_type == 'PII':
            w = self.protein.residue_PII_weights
        elif move_type == 'turn1':
            w = self.protein.residue_turn1_weights
        elif move_type == 'turn2':
            w = self.protein.residue_turn2_weights
        else:
            raise TypeError, 'Unknown move_type %s' % move_type

        weight = float(weight)
        rn = self.protein.residue_names

        for arg in res:
            if arg == 'all':
                for i in range(self.protein.num_residues):
                    w[i] = weight
            elif type(arg) is StringType:
                for i in range(self.protein.num_residues):
                    if rn[i] == arg: w[i] = weight
            elif type(arg) is IntType:
                w[arg] = weight
            else:
                raise TypeError, 'Unknown residue %s ' % arg

    def normalize_weights(self):
        """Normalize each residue's sampling weights so that it sums to 1.0

        Arguments

            None
        """
        
        p = self.protein
        hw = p.residue_helix_weights
        sw = p.residue_strand_weights
        cw = p.residue_coil_weights
        pw = p.residue_PII_weights
        t1w = p.residue_turn1_weights
        t2w = p.residue_turn2_weights
        for i in range(p.num_residues):
            try:
                norm = 1.0 / (hw[i] + sw[i] + t1w[1] + t2w[i] + cw[i])
            except ZeroDivisionError:
                pass
            else:
                hw[i] = hw[i] * norm
                sw[i] = sw[i] * norm
                cw[i] = cw[i] * norm
                pw[i] = pw[i] * norm
                t1w[i] = t1w[i] * norm
                t2w[i] = t2w[i] * norm

    def read_weights_from_file(self, filename, skip=4):
        """Read sampling weights from a file

        Arguments

            o *filename* - string - name of file containing the weights

            o *skip* - integer - number of leading lines to skip in file

        Description

            Weights for every residue should be specified.  Look at
            a SWT file output from the simulation for an example of
            a valid input file

        """    
        from string import split, atof
        data = map(split, open(filename, 'rb').readlines()[skip:])
        hw = []
        sw = []
        cw = []
        pw = []
        t1w = []
        t2w = []
        for line in data:
            h, s, t1, t2, c, p = map(atof, line[2:8])
            hw.append(h)
            sw.append(s)
            t1w.append(t1)
            t2w.append(t2)
            cw.append(c)
            pw.append(p)
        if len(hw) <> self.protein.num_residues:
            raise TypeError, 'file does not contain weights for all residues'
        self.protein.residue_helix_weights = hw
        self.protein.residue_strand_weights = sw
        self.protein.residue_turn1_weights = t1w
        self.protein.residue_turn2_weights = t2w
        self.protein.residue_coil_weights = cw
        self.protein.residue_PII_weights = pw

    def update_weights(self, W):
        """Update sampling weights from a list of weights

        Arguments

           o *W* - list of lists - each list item is a list of length
           6 containing, respectively, the helix, strand, turn1, turn2,
           coil and PII weights.

        """   
        hw = self.protein.residue_helix_weights
        sw = self.protein.residue_strand_weights
        t1w = self.protein.residue_turn1_weights
        t2w = self.protein.residue_turn2_weights
        cw = self.protein.residue_coil_weights
        pw = self.protein.residue_PII_weights

        for i in range(len(hw)):
             hw[i], sw[i], t1w[i], t2w[i], cw[i], pw[i] = W[i]

    def load_pdb_weights(self):
        """Experimental - dont use"""
        
        from linusRes import PDB_WEIGHTS
        rn = self.protein.residue_names
        hw = self.protein.residue_helix_weights
        sw = self.protein.residue_strand_weights
        t1w = self.protein.residue_turn1_weights
        t2w = self.protein.residue_turn2_weights
        cw = self.protein.residue_coil_weights
        for i in range(1, self.protein.num_residues-1):
            key = rn[i-1], rn[i], rn[i+1]
            try:
                h, s, t1, t2, c = PDB_WEIGHTS[key]
            except KeyError:
                pass
            else:
                hw[i] = h
                sw[i] = s
                t1w[i] = t1
                t2w[i] = t2
                cw[i] = c

    def total_score(self):
        """Total energy of molecule

        Arguments

            none

        Result

            The energy as a floating point number
        """
            
        etot = 0.0
        prot = self.protein
        if self.hbdpar['use_hbond'] or self.hbdpar['use_sidechain_hbond']:
            etot = hbond_score(self.hblist)
        if self.conpar['use_contact']:
            etot = etot + contact_score(self.conlist,
                                        self.conpar['contact_probe'])
        if self.torpar['use_torsion']:
            etot = etot + torsion_score(self.protein, 
                                        self.torpar['torsion_penalty'])
        if self.slvpar['use_solvation']:
            area, esolv = solvation_score(prot.satoms, self.solvation_winmax)
            etot = etot + esolv + area * self.cavity_term
            
        if self.compar['use_compact']:
            etot = etot / compaction_score(self.protein,
                                           self.compar['compact_fraction'])

        return etot

    def register_functions(self, func, when=2):
        """Register functions to be run before, during or after a simulation

        Arguments

            o *func* - a python/C function

            o *when* - integer - if 0 the function is called prior to running
            the simulation, if 1 the function is called after every cycle
            of the simulation and if 2 the function is called at the
            end of the simulation

        Description

            This function allows users to define their own analysis
            functions that can be executed at defined points in the
            simulation.  The user defined function will be called with
            one argument - the simulation object.
        """    
        self.registered_functions.append((func, when))

    def run(self):
        """Run the simulation"""
        import construct
        for i in range(self.trials):
            self.trial = i + 1
            if self.trial > 1:
                if self.reset_conformation:
                    self.protein.reset_conformation()
                if self.updateweights: 
                    self.update_weights(self.swt)
            if self.use_hbond or self.use_sidechain_hbond:
                self.hblist = construct.make_hbond_list(self.protein,
                                                    self.hbond_winmin,
                                                    self.hbond_winmax,
                                                    self.hbdpar)
            if self.conpar['use_contact']:
                self.conlist = construct.make_contact_list(self.protein,
                                                       self.contact_winmin,
                                                       self.contact_winmax,
						       self.contact_scale)
            if self.use_solvation:
                self.protein.satoms = construct.set_solvation_parameters(
                    self.protein, self.solvation_probe)

            if self.use_ms_sampling:
                from mesodict import RES_MESOSTATES
                if hasattr(self.protein, 'mesostates'):
                    del self.protein.mesostates
                ms = self.protein.mesostates = []
                for i in range(self.protein.num_residues):
                    ms.append(RES_MESOSTATES[self.protein.residue_names[i]])
                
            for func, when in self.registered_functions:
                if when == 0:
                    func(self)
                    
            linusRun.run(self)

            for func, when in self.registered_functions:
                if when == 2:
                    func(self)

        if self.reset_conformation:
            self.protein.reset_conformation()
        if self.updateweights: 
            self.update_weights(self.swt)
##        for func, when in self.registered_functions:
##            if when == 2:
##                func(self)
        self.trial = 0

    def load_mesostates(self, file):
	from string import strip
        data = map(strip, open(file, 'r').readlines())
        p = self.protein
        m = p.mesostates = []
        for d in data:
            if d[:1] == '?':
                m.append('')
            else:
                m.append(d)
